# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Functionality to animate rigs."""

import enum
import maya.api.OpenMaya as om
import puppeteer.mayax as mx
import puppeteer.rigger as rigger
import puppeteer.rigquery as rigquery
import puppeteer.rigdefinition as rigdef
import puppeteer.rigcontrol as rigcontrol
import puppeteer.rigutils as rigutils
import puppeteer.utils as utils
import puppeteer.animutils as animutils
import puppeteer.dynamics as dynamics
import puppeteer.ui as ui


class AnimatorError(Exception):
    """Raised by the Animator."""


class AnimatorMode(enum.Enum):
    """The animator's mode of operation."""

    ANIMATE = enum.auto()
    POSE = enum.auto()
    SETUP = enum.auto()


class AnimatorState(ui.BindableData):
    """The state for the Animator tool."""

    class RigControlState(ui.BindableData):
        obj = None
        parentSpace = []
        positionSpace = []
        rotationSpace = []
        scaleSpace = []
        aimSpace = []
        particle = None

    class UIState(ui.BindableData):
        rigControlPropertiesCollapsed = True
        rigControlSpacesCollapsed = False
        rigControlDynamicsCollapsed = True

    mode = AnimatorMode.ANIMATE
    selectedRigControl = RigControlState()

    uiState = UIState()

    _objSelectionChangedCallbackId = 0

    def __init__(self, **properties):
        super().__init__(**properties)

        self._objSelectionChangedCallbackId = om.MEventMessage.addEventCallback(
            'SelectionChanged',
            utils.weakMethodProxy(self._onObjectSelectionChanged),
        )
        self._onObjectSelectionChanged()

    def __del__(self):
        om.MEventMessage.removeCallback(self._objSelectionChangedCallbackId)

    def _onObjectSelectionChanged(self, *_):
        """Call on Maya object selection change."""
        selectionList = mx.cmd.ls(selection=True, objectsOnly=True)

        if selectionList and rigcontrol.isControl(selectionList[0]):
            control = selectionList[0]

            self.selectedRigControl.parentSpace = rigcontrol.getControlParentSpace(control)
            self.selectedRigControl.positionSpace = rigcontrol.getControlPositionSpace(control)
            self.selectedRigControl.rotationSpace = rigcontrol.getControlRotationSpace(control)
            self.selectedRigControl.scaleSpace = rigcontrol.getControlScaleSpace(control)
            self.selectedRigControl.aimSpace = rigcontrol.getControlAimSpace(control)
            self.selectedRigControl.particle = dynamics.getControlParticle(control)
            self.selectedRigControl.obj = control
        else:
            self.selectedRigControl.clear()


# --------------------------------------------------------------------------------------------------


@mx.undoable
def importRig(filePath, targetRoot=None):
    """Import a rig from a definition file while preserving the motion."""
    rigRoot = rigger.importRig(filePath, targetRoot)
    controls = rigquery.getAllControls(rigRoot)

    transferAnimationFromTargetsToControls(controls)

    updatePoleControlsAnimation(controls)

    return rigRoot


@mx.undoable
def applyPreset(
    rigRoot,
    preset: rigdef.RigDefinition.Preset,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Apply the rig preset while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigger.applyPreset(rigRoot, preset)
        return

    controls = []

    for control in preset.controls:
        try:
            controls.append(mx.Node(f'{rigRoot.pathName}|{control.path}'))
        except mx.MayaNodeError:
            continue

    preserveKeys = animMode is AnimatorMode.POSE

    with animutils.preserveMotion(controls, preserveKeys=preserveKeys):
        rigger.applyPreset(rigRoot, preset)

    updatePoleControlsAnimation(controls, preserveKeys=preserveKeys)


@mx.undoable
def updatePoleControlsAnimation(controls, preserveKeys=False):
    """Update the pole controls' position each frame to follow animated chain."""
    ikChains = []
    ikControls = []

    for control in controls:
        ikChain = rigquery.getIKChainFromSingleControl(control)

        if ikChain and ikChain not in ikChains:
            ikChains.append(ikChain)
            ikControls.extend(ikChain.controls)

    with animutils.preserveMotion(ikControls, preserveKeys=preserveKeys) as bakedLocators:
        for ikChain in ikChains:
            ikChainIndex = ikControls.index(ikChain.controls[0])
            ikChainLocators = bakedLocators[ikChainIndex : ikChainIndex + len(ikChain.controls)]

            transferAnimationFromLocatorsToPoleControl(ikChainLocators, ikChain.poleControl)

            if preserveKeys:
                keysToPreserve = animutils.getFramesWithTransformKeys(ikChain.controls[-1])

                if keysToPreserve:
                    keysToRemove = [
                        (key,)
                        for key in animutils.getFramesWithTransformKeys(ikChain.poleControl)
                        if key not in keysToPreserve
                    ]
                    mx.cmd.cutKey(ikChain.poleControl, time=keysToRemove, clear=True)


@mx.undoable
def transferAnimationFromTargetsToControls(controls):
    """Transfer animation from targets to rig controls."""
    bakedControls = []
    tempNodes = []

    for control in controls:
        if control.target and control.constraint.type != 'poleVectorConstraint':
            rigcontrol.disconnectControl(control)

            bakedControls.append(control)

    if not bakedControls:
        return

    mx.cmd.cutKey(bakedControls, clear=True, time=(), attribute=['translate', 'rotate', 'scale'])

    for control in bakedControls:
        parentConstraint = mx.cmd.parentConstraint(control.target, control, maintainOffset=True)[0]
        scaleConstraint = mx.cmd.scaleConstraint(control.target, control, maintainOffset=True)[0]

        # TODO: Remove if the rig control is no longer using the joint for its transform node.
        scaleConstraint['constraintScaleCompensate'].disconnectInput()
        scaleConstraint.constraintScaleCompensate = False

        scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
        scaleConstraint['constraintScale'].connect(scaleOffsetNode['input1'])
        control['inverseScale'].connect(scaleOffsetNode['input2'])

        scaleOffsetNode['outputX'].connect(control['scaleX'], force=True)
        scaleOffsetNode['outputY'].connect(control['scaleY'], force=True)
        scaleOffsetNode['outputZ'].connect(control['scaleZ'], force=True)

        tempNodes.append(parentConstraint)
        tempNodes.append(scaleConstraint)
        tempNodes.append(scaleOffsetNode)

    animutils.bake(bakedControls)

    mx.cmd.delete(tempNodes)

    for control in bakedControls:
        rigcontrol.reconnectControl(control)

        mx.cmd.cutKey(control.target, clear=True, time=())


@mx.undoable
def transferAnimationFromLocatorsToPoleControl(locatorsChain, poleControl, timeRange=None):
    """Transfer animation from locators to pole vector control."""
    if not timeRange:
        # TODO: Use a function to get the entire range.
        timeRange = (
            mx.cmd.playbackOptions(query=True, animationStartTime=True),
            mx.cmd.playbackOptions(query=True, animationEndTime=True),
        )

    position, rotation = rigutils.calculatePoleVectorPositionAndRotation(
        locatorsChain,
        time=timeRange[0],
    )
    poleControl.worldPosition = position
    poleControl.worldRotation = rotation

    for time in range(int(timeRange[0]), int(timeRange[1]) + 1):
        polePosition, poleRotation = rigutils.calculatePoleVectorPositionAndRotation(
            locatorsChain,
            objSpace=poleControl,
            time=time,
        )

        mx.cmd.setKeyframe(poleControl, attribute='translateX', value=polePosition.x, time=time)
        mx.cmd.setKeyframe(poleControl, attribute='translateY', value=polePosition.y, time=time)
        mx.cmd.setKeyframe(poleControl, attribute='translateZ', value=polePosition.z, time=time)

        mx.cmd.setKeyframe(poleControl, attribute='rotateX', value=poleRotation.x, time=time)
        mx.cmd.setKeyframe(poleControl, attribute='rotateY', value=poleRotation.y, time=time)
        mx.cmd.setKeyframe(poleControl, attribute='rotateZ', value=poleRotation.z, time=time)


@mx.undoable
def transferAnimationBetweenSelectedObjects():
    """Transfer animation between the selected objects."""
    selectionList = mx.cmd.ls(selection=True)

    if len(selectionList) != 2:
        raise AnimatorError('Two objects should be selected.')

    animutils.transferAnimation(*selectionList)


@mx.undoable
def transferPositionAnimationBetweenSelectedObjects():
    """Transfer position animation between the selected objects."""
    selectionList = mx.cmd.ls(selection=True)

    if len(selectionList) != 2:
        raise AnimatorError('Two objects should be selected.')

    animutils.transferPositionAnimation(*selectionList)


@mx.undoable
def transferRotationAnimationBetweenSelectedObjects():
    """Transfer rotation animation between the selected objects."""
    selectionList = mx.cmd.ls(selection=True)

    if len(selectionList) != 2:
        raise AnimatorError('Two objects should be selected.')

    animutils.transferRotationAnimation(*selectionList)


@mx.undoable
def transferScaleAnimationBetweenSelectedObjects():
    """Transfer scale animation between the selected objects."""
    selectionList = mx.cmd.ls(selection=True)

    if len(selectionList) != 2:
        raise AnimatorError('Two objects should be selected.')

    animutils.transferScaleAnimation(*selectionList)


# --------------------------------------------------------------------------------------------------


@mx.undoable
def deleteControls(controls):
    """Delete the rig controls and transfer the motion to their targets."""
    objectsToPreserve = [control.target for control in controls if control.target]
    controlsToPreserve = []

    for ctrl in controls:
        ikChain = rigquery.getIKChainFromSingleControl(ctrl)

        if ikChain:
            controlsToPreserve.extend(
                [
                    ikCtrl
                    for ikCtrl in ikChain.controls
                    if ikCtrl not in controls
                    if ikCtrl not in controlsToPreserve
                ]
            )

    bakedLocators = animutils.transferAnimationToLocators(objectsToPreserve, relativeToObject=True)

    with animutils.preserveMotion(controlsToPreserve):
        rigger.deleteControls(controls)

    objectsToPreserve = [obj for obj in objectsToPreserve if obj.exists]
    bakedLocators = [locator for locator in bakedLocators if locator.exists]

    for i, obj in enumerate(objectsToPreserve):
        locator = bakedLocators[i]

        mx.cmd.cutKey(obj, clear=True, time=(), attribute=['translate', 'rotate', 'scale'])

        animTranslateX = locator['translateX'].input()
        animTranslateX.name = f'{obj.name}_translateX'
        animTranslateX['output'].connect(obj['translateX'])

        animTranslateY = locator['translateY'].input()
        animTranslateY.name = f'{obj.name}_translateY'
        animTranslateY['output'].connect(obj['translateY'])

        animTranslateZ = locator['translateZ'].input()
        animTranslateZ.name = f'{obj.name}_translateZ'
        animTranslateZ['output'].connect(obj['translateZ'])

        animRotateX = locator['rotateX'].input()
        animRotateX.name = f'{obj.name}_rotateX'
        animRotateX['output'].connect(obj['rotateX'])

        animRotateY = locator['rotateY'].input()
        animRotateY.name = f'{obj.name}_rotateY'
        animRotateY['output'].connect(obj['rotateY'])

        animRotateZ = locator['rotateZ'].input()
        animRotateZ.name = f'{obj.name}_rotateZ'
        animRotateZ['output'].connect(obj['rotateZ'])

        animScaleX = locator['scaleX'].input()
        animScaleX.name = f'{obj.name}_scaleX'
        animScaleX['output'].connect(obj['scaleX'])

        animScaleY = locator['scaleY'].input()
        animScaleY.name = f'{obj.name}_scaleY'
        animScaleY['output'].connect(obj['scaleY'])

        animScaleZ = locator['scaleZ'].input()
        animScaleZ.name = f'{obj.name}_scaleZ'
        animScaleZ['output'].connect(obj['scaleZ'])

    if bakedLocators:
        mx.cmd.delete(bakedLocators)


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlParentSpace(
    control,
    target,
    operation=rigcontrol.SpaceOperation.REPLACE,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Change the rig control's parent space while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigcontrol.changeControlParentSpace(control, target, operation)
    else:
        with animutils.preserveMotion([control], preserveKeys=animMode is AnimatorMode.POSE):
            rigcontrol.changeControlParentSpace(control, target, operation)

    mx.cmd.select(control)


@mx.undoable
def changeControlPositionSpace(
    control,
    target,
    operation=rigcontrol.SpaceOperation.REPLACE,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Change the rig control's position space while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigcontrol.changeControlPositionSpace(control, target, operation)
    else:
        with animutils.preserveMotion([control], preserveKeys=animMode is AnimatorMode.POSE):
            rigcontrol.changeControlPositionSpace(control, target, operation)

    mx.cmd.select(control)


@mx.undoable
def changeControlRotationSpace(
    control,
    target,
    operation=rigcontrol.SpaceOperation.REPLACE,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Change the rig control's rotation space while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigcontrol.changeControlRotationSpace(control, target, operation)
    else:
        with animutils.preserveMotion([control], preserveKeys=animMode is AnimatorMode.POSE):
            rigcontrol.changeControlRotationSpace(control, target, operation)

    mx.cmd.select(control)


@mx.undoable
def changeControlScaleSpace(
    control,
    target,
    operation=rigcontrol.SpaceOperation.REPLACE,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Change the rig control's scale space while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigcontrol.changeControlScaleSpace(control, target, operation)
    else:
        with animutils.preserveMotion([control], preserveKeys=animMode is AnimatorMode.POSE):
            rigcontrol.changeControlScaleSpace(control, target, operation)

    mx.cmd.select(control)


@mx.undoable
def changeControlAimSpace(
    control,
    target,
    operation=rigcontrol.SpaceOperation.REPLACE,
    animMode: AnimatorMode = AnimatorMode.ANIMATE,
):
    """Change the rig control's aim space while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigcontrol.changeControlAimSpace(control, target, operation)
    else:
        with animutils.preserveMotion([control], preserveKeys=animMode is AnimatorMode.POSE):
            rigcontrol.changeControlAimSpace(control, target, operation)

    mx.cmd.select(control)


# --------------------------------------------------------------------------------------------------


@mx.undoable
def switchControlsToFK(controls, animMode: AnimatorMode = AnimatorMode.ANIMATE):
    """Switch controls to FK while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigger.switchControlsToFK(controls)
        return

    if len(controls) == 1:
        ikChain = rigquery.getIKChainFromSingleControl(controls[0])

        if ikChain:
            controls = ikChain.controls

    with animutils.preserveMotion(controls, preserveKeys=animMode is AnimatorMode.POSE):
        rigger.switchControlsToFK(controls)


@mx.undoable
def switchControlsToReversedFK(controls, animMode: AnimatorMode = AnimatorMode.ANIMATE):
    """Switch controls to Reversed FK while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        rigger.switchControlsToReversedFK(controls)
        return

    if len(controls) == 1:
        ikChain = rigquery.getIKChainFromSingleControl(controls[0])

        if ikChain:
            controls = ikChain.controls
        else:
            controls = rigquery.getFKChainFromSingleControl(controls[0])

    with animutils.preserveMotion(controls, preserveKeys=animMode is AnimatorMode.POSE):
        rigger.switchControlsToReversedFK(controls)


@mx.undoable
def switchControlsToIK(controls, animMode: AnimatorMode = AnimatorMode.ANIMATE):
    """Switch controls to IK while preserving the motion."""
    if animMode is AnimatorMode.SETUP:
        return rigger.switchControlsToIK(controls)

    if len(controls) == 1:
        controls = rigquery.getFKChainFromSingleControl(controls[0])

    with animutils.preserveMotion(
        controls,
        preserveKeys=animMode is AnimatorMode.POSE,
    ) as bakedLocators:
        ikChain = rigger.switchControlsToIK(controls)

        if ikChain:
            transferAnimationFromLocatorsToPoleControl(bakedLocators, ikChain.poleControl)

            if animMode is AnimatorMode.POSE:
                keysToPreserve = animutils.getFramesWithTransformKeys(ikChain.controls[-1])

                if keysToPreserve:
                    keysToRemove = [
                        (key,)
                        for key in animutils.getFramesWithTransformKeys(ikChain.poleControl)
                        if key not in keysToPreserve
                    ]
                    mx.cmd.cutKey(ikChain.poleControl, time=keysToRemove, clear=True)

        return ikChain


# --------------------------------------------------------------------------------------------------


@mx.undoable
def createControlsForSelectedObjects():
    """Create rig controls for the selected objects while preserving the motion."""
    controls = rigcontrol.createControlsForSelectedObjects()

    transferAnimationFromTargetsToControls(controls)


@mx.undoable
def keyRig(rigRoot):
    """Key the rig controls."""
    controls = rigquery.getAllControls(rigRoot)

    if controls:
        mx.cmd.setKeyframe(controls)


@mx.undoable
def keySelectedRig():
    """Key the rig controls for the selected rig(s)."""
    rigs = []

    for control in rigcontrol.getSelectedControls():
        rigRoot = rigquery.getRigRootFromControl(control)

        if rigRoot and rigRoot not in rigs:
            rigs.append(rigRoot)

    for rigRoot in rigs:
        keyRig(rigRoot)


@mx.undoable
def bakeRig(rigRoot, timeRange=None):
    """Bake the rig controls."""
    controls = rigquery.getAllControls(rigRoot)

    if controls:
        animutils.bake(controls, timeRange=timeRange)


@mx.undoable
def bakeSelectedRig(timeRange=None):
    """Bake the rig controls for the selected rig(s)."""
    rigs = []

    for control in rigcontrol.getSelectedControls():
        rigRoot = rigquery.getRigRootFromControl(control)

        if rigRoot and rigRoot not in rigs:
            rigs.append(rigRoot)

    for rigRoot in rigs:
        bakeRig(rigRoot, timeRange=timeRange)
