# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Constraints using matrix connections."""

import puppeteer.mayax as mx


@mx.undoable
def parentConstraint(parent, child, maintainOffset=True):
    """Create a parent constraint using matrix connections."""
    if not maintainOffset:
        mx.cmd.matchTransform(child, parent)

    constraintName = f'{parent.name}__{child.name}__mtxParentConstraint'
    multiplyMatrixNode = mx.cmd.createNode('multMatrix', name=constraintName, skipSelect=True)

    offset = child.parentMatrix * parent.worldInverseMatrix
    multiplyMatrixNode['matrixIn[0]'].value = offset

    parent['worldMatrix'].connect(multiplyMatrixNode['matrixIn[1]'])

    childParent = child.parent
    if childParent and child.inheritsTransform:
        childParent['worldInverseMatrix'].connect(multiplyMatrixNode['matrixIn[2]'])

    multiplyMatrixNode['matrixSum'].connect(child['offsetParentMatrix'], force=True)

    multiplyMatrixNode.addAttr(
        '__mtxParentConstraint__',
        True,
        niceName='__mtxParentConstraint__',
    ).locked = True

    return multiplyMatrixNode


@mx.undoable
def deleteParentConstraints(obj, otherObj=None, parentOnly=False, childsOnly=False):
    """Delete the matrix parent constraints for the given object.

    If `otherObj` is not provided, all constraints will be removed.
    """
    constraints = getParentConstraints(obj, otherObj, parentOnly, childsOnly)

    if constraints:
        tempTransform = mx.cmd.createNode(
            'transform',
            name=f'{obj}_mtxTempTransform',
            skipSelect=True,
        )

        for constraint in constraints:
            for child in constraint['matrixSum'].outputs():
                mx.cmd.matchTransform(tempTransform, child)

                child['offsetParentMatrix'].disconnectInput()
                child.offsetParentMatrix = mx.Matrix.kIdentity

                mx.cmd.matchTransform(child, tempTransform)

        mx.cmd.delete(constraints)
        mx.cmd.delete(tempTransform)


def getParentConstraints(obj, otherObj=None, parentOnly=False, childsOnly=False, target=False):
    """Retrieve the matrix parent constraints (or their targets) for the given object."""
    result = []

    if parentOnly or not childsOnly:
        constraint = obj['offsetParentMatrix'].input(type='multMatrix')

        if constraint and not constraint.hasAttr('__mtxParentConstraint__'):
            constraint = None

        if constraint:
            targetObj = constraint['matrixIn[1]'].input() if otherObj or target else None

            if not otherObj or otherObj == targetObj:
                if not target:
                    result.append(constraint)
                elif targetObj:
                    result.append(targetObj)

    if not parentOnly or childsOnly:
        childConstraints = [
            constraint
            for constraint in obj['worldMatrix'].outputs(type='multMatrix')
            if (
                constraint.hasAttr('__mtxParentConstraint__')
                and (not otherObj or otherObj in constraint['matrixSum'].outputs())
            )
        ]

        if not target:
            result.extend(childConstraints)
        else:
            result.extend(
                [
                    targetObj
                    for constraint in childConstraints
                    for targetObj in constraint['matrixSum'].outputs()
                ]
            )

    return result
