# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Rig control's spaces functionality."""

import puppeteer.mayax as mx
import puppeteer.rigutils as rigutils
from .main import isControl
from .links import LinkTarget, LinkDirection, linkControl, unlinkControl


class SpaceOperation:
    """Rig control's space switch operation type."""

    ADD = 1
    REMOVE = 2
    REPLACE = 3
    UPDATE = 4


class SpaceTarget:
    """Rig control's space target."""

    def __init__(self, target, weight=1.0):
        self.target = target
        self.weight = weight

    def __eq__(self, other):
        return self.target == other

    def __ne__(self, other):
        return self.target != other


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlParentSpace(control, target, operation=SpaceOperation.REPLACE):
    """Constrain the rig control's position, rotation, and scale to the provided target."""
    worldMatrix = control.worldMatrix
    blendNode = control['parentBuffer'].input()

    if isinstance(target, SpaceTarget):
        space = target
        target = space.target
    else:
        space = SpaceTarget(target)

    if (
        operation in (SpaceOperation.REPLACE, SpaceOperation.ADD)
        and isControl(target)
        and control in getControlParentSpace(target)
    ):
        # avoid evaluation cycles
        changeControlParentSpace(target, control, SpaceOperation.REMOVE)

    # remove control links
    if not target or operation == SpaceOperation.REPLACE:
        for oldSpace in getControlParentSpace(control):
            unlinkControl(control, oldSpace.target)
    elif operation == SpaceOperation.REMOVE:
        unlinkControl(control, target)

    if blendNode:
        if operation == SpaceOperation.REPLACE:
            blendNode.delete()
            blendNode = None
        elif operation == SpaceOperation.REMOVE or (
            operation == SpaceOperation.ADD and target in getControlParentSpace(control)
        ):
            existingSpaces = getControlParentSpace(control)

            if len(existingSpaces) == 1:
                blendNode.delete()
                blendNode = None
            else:
                targetIndices = mx.cmd.getAttr(blendNode['target'], multiIndices=True)
                targetIndex = existingSpaces.index(target)
                targetAttr = f'target[{targetIndices[targetIndex]}]'

                target['worldMatrix'].disconnect(blendNode[f'{targetAttr}.targetMatrix'])

                # Update weight values to be weighted average.
                del targetIndices[targetIndex]
                for index, logicalIndex in enumerate(targetIndices):
                    targetAttr = f'target[{logicalIndex}]'
                    blendNode[f'{targetAttr}.weight'].value = 1.0 / (index + 1)

    existingSpaces = getControlParentSpace(control)

    if not existingSpaces:
        control.parentBuffer = mx.Matrix.kIdentity

    if target and operation != SpaceOperation.REMOVE:
        if not blendNode and operation != SpaceOperation.UPDATE:
            blendNode = mx.cmd.createNode(
                'blendMatrix',
                name=f'{control.name}__parentSpaceBlend',
                skipSelect=True,
            )
            blendNode['outputMatrix'].connect(control['parentBuffer'])

        targetIndices = mx.cmd.getAttr(blendNode['target'], multiIndices=True)

        if operation == SpaceOperation.UPDATE:
            targetIndex = targetIndices[existingSpaces.index(target)]
        else:
            targetIndex = targetIndices[-1] + 1 if targetIndices else 0

        targetAttr = f'target[{targetIndex}]'

        if operation != SpaceOperation.UPDATE:
            if space.weight == 1.0:  # Assume no weight was passed.
                # Calculate weighted average since blendMatrix node uses an ordered blend.
                space.weight = 1.0 if not existingSpaces else 1.0 / (len(existingSpaces) + 1)

            target['worldMatrix'].connect(blendNode[f'{targetAttr}.targetMatrix'])

            linkControl(
                control,
                LinkTarget(
                    target,
                    color=control.color,
                    direction=LinkDirection.BACKWARD,
                ),
            )

        blendNode[f'{targetAttr}.weight'].value = space.weight

    # Setting the world matrix two times is needed. See the following tests:
    # test_should_changeControlParentSpace_and_maintainOffset_given_controlWithAimSpace
    # test_should_switchControlsToIK_and_maintainTransforms_when_changingRootParentSpace
    # TODO: Do a better fix.
    control.worldMatrix = worldMatrix
    control.worldMatrix = worldMatrix


def getControlParentSpace(control):
    """Get the target objects for the rig control's parent space."""
    blendNode = control['parentBuffer'].input()
    spaces = []

    if not blendNode:
        return spaces

    targetIndices = mx.cmd.getAttr(blendNode['target'], multiIndices=True)

    for index in targetIndices:
        targetAttr = f'target[{index}]'
        target = blendNode[f'{targetAttr}.targetMatrix'].input()

        if target:
            spaces.append(SpaceTarget(target, weight=blendNode[f'{targetAttr}.weight'].value))

    return spaces


def getControlsDrivenByParentSpace(control):
    """Get the controls driven by the parent space."""
    drivenControls = []

    for node in control['worldMatrix'].outputs():
        if node.type == 'blendMatrix':
            try:
                drivenControls.append(node['outputMatrix'].outputs()[0])
            except IndexError:
                pass

    return drivenControls


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlPositionSpace(control, target, operation=SpaceOperation.REPLACE):
    """Constrain the rig control's position to the provided target."""
    worldPosition = control.worldPosition
    constraint = control['translateBuffer'].input()

    if isinstance(target, SpaceTarget):
        space = target
        target = space.target
    else:
        space = SpaceTarget(target)

    if (
        (operation == SpaceOperation.REPLACE or operation == SpaceOperation.ADD)
        and isControl(target)
        and control in getControlPositionSpace(target)
    ):
        # avoid evaluation cycles
        changeControlPositionSpace(target, control, SpaceOperation.REMOVE)

    if constraint:
        if operation == SpaceOperation.REPLACE:
            constraint.delete()
            constraint = None
        elif operation == SpaceOperation.REMOVE or (
            operation == SpaceOperation.ADD and target in getControlPositionSpace(control)
        ):
            if not mx.cmd.pointConstraint(target, constraint, remove=True):
                # Target was the only entry and the constraint was removed.
                constraint = None

    existingSpaces = getControlPositionSpace(control)

    if not existingSpaces:
        control.translateBuffer = mx.Vector.kZeroVector

    if target and operation != SpaceOperation.REMOVE:
        if not constraint and operation != SpaceOperation.UPDATE:
            constraint = mx.cmd.createNode(
                'pointConstraint',
                name='positionSpaceConstraint',
                parent=control,
                skipSelect=True,
            )
            constraint.hiddenInOutliner = True
            constraint['constraintTranslate'].connect(control['translateBuffer'])
            control['rotatePivot'].connect(constraint['constraintRotatePivot'])
            control['rotatePivotTranslate'].connect(constraint['constraintRotateTranslate'])

        targetIndices = mx.cmd.getAttr(constraint['target'], multiIndices=True)

        if operation == SpaceOperation.UPDATE:
            targetIndex = targetIndices[existingSpaces.index(target)]
        else:
            targetIndex = targetIndices[-1] + 1 if targetIndices else 0

        targetAttr = f'target[{targetIndex}]'

        if operation != SpaceOperation.UPDATE:
            target['translate'].connect(constraint[f'{targetAttr}.targetTranslate'])
            target['rotatePivot'].connect(constraint[f'{targetAttr}.targetRotatePivot'])
            target['rotatePivotTranslate'].connect(
                constraint[f'{targetAttr}.targetRotateTranslate']
            )
            target['parentMatrix'].connect(constraint[f'{targetAttr}.targetParentMatrix'])

        constraint[f'{targetAttr}.targetWeight'].value = space.weight

    control.worldPosition = worldPosition


def getControlPositionSpace(control):
    """Get the target objects for the rig control's position space."""
    constraint = control['translateBuffer'].input()
    spaces = []

    if not constraint:
        return spaces

    targets = mx.cmd.pointConstraint(constraint, query=True, targetList=True) or []

    for target in targets:
        spaces.append(
            SpaceTarget(
                target,
                weight=mx.cmd.pointConstraint(target, constraint, query=True, weight=True),
            )
        )

    return spaces


def getControlsDrivenByPositionSpace(control):
    """Get the controls driven by the position space."""
    drivenControls = []

    for node in control['translate'].outputs():
        if node.type == 'pointConstraint':
            try:
                drivenControls.append(node['constraintTranslate'].outputs()[0])
            except IndexError:
                continue

    return drivenControls


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlRotationSpace(control, target, operation=SpaceOperation.REPLACE):
    """Constrain the rig control's rotation to the provided target."""
    if getControlAimSpace(control) and operation in (
        SpaceOperation.REPLACE,
        SpaceOperation.ADD,
    ):
        changeControlAimSpace(control, None)

    worldRotation = control.worldRotation
    constraint = control['rotateBuffer'].input()

    if isinstance(target, SpaceTarget):
        space = target
        target = space.target
    else:
        space = SpaceTarget(target)

    if (
        (operation == SpaceOperation.REPLACE or operation == SpaceOperation.ADD)
        and isControl(target)
        and control in getControlRotationSpace(target)
    ):
        # avoid evaluation cycles
        changeControlRotationSpace(target, control, SpaceOperation.REMOVE)

    if constraint:
        if operation == SpaceOperation.REPLACE:
            constraint.delete()
            constraint = None
        elif operation == SpaceOperation.REMOVE or (
            operation == SpaceOperation.ADD and target in getControlRotationSpace(control)
        ):
            if not mx.cmd.orientConstraint(target, constraint, remove=True):
                # Target was the only entry and the constraint was removed.
                constraint = None

    existingSpaces = getControlRotationSpace(control)

    if not existingSpaces:
        control.rotateBuffer = mx.Vector.kZeroVector

    if target and operation != SpaceOperation.REMOVE:
        if not constraint and operation != SpaceOperation.UPDATE:
            constraint = mx.cmd.createNode(
                'orientConstraint',
                name='orientSpaceConstraint',
                parent=control,
                skipSelect=True,
            )
            constraint.hiddenInOutliner = True
            constraint['constraintRotate'].connect(control['rotateBuffer'])
            control['rotateOrder'].connect(constraint['constraintRotateOrder'])
            control['parentInverseMatrix'].connect(constraint['constraintParentInverseMatrix'])

        targetIndices = mx.cmd.getAttr(constraint['target'], multiIndices=True)

        if operation == SpaceOperation.UPDATE:
            targetIndex = targetIndices[existingSpaces.index(target)]
        else:
            targetIndex = targetIndices[-1] + 1 if targetIndices else 0

        targetAttr = f'target[{targetIndex}]'

        if operation != SpaceOperation.UPDATE:
            target['rotate'].connect(constraint[f'{targetAttr}.targetRotate'])
            target['rotateOrder'].connect(constraint[f'{targetAttr}.targetRotateOrder'])
            target['parentMatrix'].connect(constraint[f'{targetAttr}.targetParentMatrix'])

            if target.type == 'joint':
                target['jointOrient'].connect(constraint[f'{targetAttr}.targetJointOrient'])

        constraint[f'{targetAttr}.targetWeight'].value = space.weight

    control.worldRotation = worldRotation


def getControlRotationSpace(control):
    """Get the target objects for the rig control's rotation space."""
    constraint = control['rotateBuffer'].input()
    spaces = []

    if not constraint:
        return spaces

    targets = mx.cmd.orientConstraint(constraint, query=True, targetList=True) or []

    for target in targets:
        spaces.append(
            SpaceTarget(
                target,
                weight=mx.cmd.orientConstraint(target, constraint, query=True, weight=True),
            )
        )

    return spaces


def getControlsDrivenByRotationSpace(control):
    """Get the controls driven by the rotation space."""
    drivenControls = []

    for node in control['rotate'].outputs():
        if node.type == 'orientConstraint':
            try:
                drivenControls.append(node['constraintRotate'].outputs()[0])
            except IndexError:
                continue

    return drivenControls


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlScaleSpace(control, target, operation=SpaceOperation.REPLACE):
    """Constrain the rig control's scale to the provided target."""
    worldScale = control.worldScale
    constraint = control['scaleBuffer'].input()

    if isinstance(target, SpaceTarget):
        space = target
        target = space.target
    else:
        space = SpaceTarget(target)

    if (
        (operation == SpaceOperation.REPLACE or operation == SpaceOperation.ADD)
        and isControl(target)
        and control in getControlScaleSpace(target)
    ):
        # avoid evaluation cycles
        changeControlScaleSpace(target, control, SpaceOperation.REMOVE)

    if constraint:
        if operation == SpaceOperation.REPLACE:
            constraint.delete()
            constraint = None
        elif operation == SpaceOperation.REMOVE or (
            operation == SpaceOperation.ADD and target in getControlScaleSpace(control)
        ):
            if not mx.cmd.scaleConstraint(target, constraint, remove=True):
                # Target was the only entry and the constraint was removed.
                constraint = None

    existingSpaces = getControlScaleSpace(control)

    if not existingSpaces:
        control.scaleBuffer = mx.Vector.kOneVector

    if target and operation != SpaceOperation.REMOVE:
        if not constraint and operation != SpaceOperation.UPDATE:
            constraint = mx.cmd.createNode(
                'scaleConstraint',
                name='scaleSpaceConstraint',
                parent=control,
                skipSelect=True,
            )
            constraint.hiddenInOutliner = True
            constraint['constraintScale'].connect(control['scaleBuffer'])
            control['parentInverseMatrix'].connect(constraint['constraintParentInverseMatrix'])

        targetIndices = mx.cmd.getAttr(constraint['target'], multiIndices=True)

        if operation == SpaceOperation.UPDATE:
            targetIndex = targetIndices[existingSpaces.index(target)]
        else:
            targetIndex = targetIndices[-1] + 1 if targetIndices else 0

        targetAttr = f'target[{targetIndex}]'

        if operation != SpaceOperation.UPDATE:
            target['scale'].connect(constraint[f'{targetAttr}.targetScale'])
            target['parentMatrix'].connect(constraint[f'{targetAttr}.targetParentMatrix'])

        constraint[f'{targetAttr}.targetWeight'].value = space.weight

    control.worldScale = worldScale


def getControlScaleSpace(control):
    """Get the target objects for the rig control's scale space."""
    constraint = control['scaleBuffer'].input()
    spaces = []

    if not constraint:
        return spaces

    targets = mx.cmd.scaleConstraint(constraint, query=True, targetList=True) or []

    for target in targets:
        spaces.append(
            SpaceTarget(
                target,
                weight=mx.cmd.scaleConstraint(target, constraint, query=True, weight=True),
            )
        )

    return spaces


def getControlsDrivenByScaleSpace(control):
    """Get the controls driven by the scale space."""
    drivenControls = []

    for node in control['scale'].outputs():
        if node.type == 'scaleConstraint':
            try:
                drivenControls.append(node['constraintScale'].outputs()[0])
            except IndexError:
                continue

    return drivenControls


# --------------------------------------------------------------------------------------------------


@mx.undoable
def changeControlAimSpace(control, target, operation=SpaceOperation.REPLACE):
    """Constrain the rig control's orientation to point at the provided target."""
    if getControlRotationSpace(control) and operation in (
        SpaceOperation.REPLACE,
        SpaceOperation.ADD,
    ):
        changeControlRotationSpace(control, None)

    worldRotation = control.worldRotation
    constraint = control['rotateBuffer'].input()

    if isinstance(target, SpaceTarget):
        space = target
        target = space.target
    else:
        space = SpaceTarget(target)

    if constraint:
        if operation == SpaceOperation.REPLACE:
            constraint.delete()
            constraint = None
        elif operation == SpaceOperation.REMOVE or (
            operation == SpaceOperation.ADD and target in getControlAimSpace(control)
        ):
            if not mx.cmd.aimConstraint(target, constraint, remove=True):
                # Target was the only entry and the constraint was removed.
                constraint = None

            control.worldRotation = worldRotation  # Needed to properly calculate the aim vectors.

    existingSpaces = getControlAimSpace(control)

    if not existingSpaces:
        control.rotateBuffer = mx.Vector.kZeroVector
        control.worldRotation = worldRotation  # Needed to properly calculate the aim vectors.

    if target and operation != SpaceOperation.REMOVE:
        if not constraint and operation != SpaceOperation.UPDATE:
            constraint = mx.cmd.createNode(
                'aimConstraint',
                name='aimSpaceConstraint',
                parent=control,
                skipSelect=True,
            )
            constraint.worldUpType = 2  # object rotation up
            constraint.hiddenInOutliner = True
            constraint['constraintRotate'].connect(control['rotateBuffer'])

            control['translate'].connect(constraint['constraintTranslate'])
            control['rotateOrder'].connect(constraint['constraintRotateOrder'])
            control['rotatePivot'].connect(constraint['constraintRotatePivot'])
            control['rotatePivotTranslate'].connect(constraint['constraintRotateTranslate'])
            control['parentInverseMatrix'].connect(constraint['constraintParentInverseMatrix'])

        targetIndices = mx.cmd.getAttr(constraint['target'], multiIndices=True)

        if operation == SpaceOperation.UPDATE:
            targetIndex = targetIndices[existingSpaces.index(target)]
        else:
            targetIndex = targetIndices[-1] + 1 if targetIndices else 0

        targetAttr = f'target[{targetIndex}]'

        if operation != SpaceOperation.UPDATE:
            aimVector, upVector, worldUpVector = rigutils.calculateAimConstraintVectors(
                control,
                target,
            )
            constraint.aimVector = aimVector
            constraint.upVector = upVector
            constraint.worldUpVector = worldUpVector
            constraint.worldUpObject = target
            target['worldMatrix'].connect(constraint['worldUpMatrix'], force=True)

            target['translate'].connect(constraint[f'{targetAttr}.targetTranslate'])
            target['rotatePivot'].connect(constraint[f'{targetAttr}.targetRotatePivot'])
            target['rotatePivotTranslate'].connect(
                constraint[f'{targetAttr}.targetRotateTranslate']
            )
            target['parentMatrix'].connect(constraint[f'{targetAttr}.targetParentMatrix'])

        constraint[f'{targetAttr}.targetWeight'].value = space.weight
    elif target and constraint and target == constraint['worldUpMatrix'].input():
        # Recalculate aim using the remaining last target after removal.
        lastTarget = existingSpaces[-1].target
        aimVector, upVector, worldUpVector = rigutils.calculateAimConstraintVectors(
            control,
            lastTarget,
        )
        constraint.aimVector = aimVector
        constraint.upVector = upVector
        constraint.worldUpVector = worldUpVector
        constraint.worldUpObject = lastTarget
        lastTarget['worldMatrix'].connect(constraint['worldUpMatrix'], force=True)

    control.worldRotation = worldRotation


def getControlAimSpace(control):
    """Get the target objects for the rig control's aim space."""
    constraint = control['rotateBuffer'].input()
    spaces = []

    if not constraint:
        return spaces

    targets = mx.cmd.aimConstraint(constraint, query=True, targetList=True) or []

    for target in targets:
        spaces.append(
            SpaceTarget(
                target,
                weight=mx.cmd.aimConstraint(target, constraint, query=True, weight=True),
            )
        )

    return spaces


def getControlsDrivenByAimSpace(control):
    """Get the controls driven by the aim space."""
    drivenControls = []

    for node in control['translate'].outputs():
        if node.type == 'aimConstraint':
            try:
                drivenControl = node['constraintRotate'].outputs()[0]
            except IndexError:
                continue

            if drivenControl != control:
                # The control's translate attribute gets connected to constraintTranslate,
                # so the control itself gets returned as well.
                drivenControls.append(drivenControl)

    return drivenControls
