# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Functionality to create rigs."""

import re
import puppeteer.mayax as mx
import puppeteer.rigcontrol as rigcontrol
import puppeteer.rigquery as rigquery
import puppeteer.rigutils as rigutils
import puppeteer.rigdefinition as rigdef


class RiggerError(Exception):
    """Raised by the Rigger."""


# --------------------------------------------------------------------------------------------------


def exportRig(rigRoot, filePath):
    """Export a rig to a definition file."""
    rigDefinition = rigdef.RigDefinition()
    targetRootPath = None

    if rigRoot.type != 'transform':
        raise RiggerError('The provided object is not a group.')

    rigRootPath = rigRoot.pathName

    for control in rigquery.getAllControls(rigRoot):
        target = control.target

        if control.constraint and control.constraint.type == 'poleVectorConstraint':
            continue

        if target and control:
            if targetRootPath is None:
                targetRootPath = f'|{target.pathName.split("|")[1]}'
            elif not target.pathName.startswith(targetRootPath):
                raise RiggerError('All rig controls should have their targets under the same root.')

        rigDefinition.controls.append(
            rigdef.RigDefinition.Control(
                path=getObjectPathRelativeToRoot(control, rigRootPath),
                targetPath=getObjectPathRelativeToRoot(target, targetRootPath) if target else None,
                position=control.worldPosition,
                rotation=control.worldRotation,
                scale=control.worldScale,
                shapeType=control.shapeType,
                size=control.size,
                innerSize=control.innerSize,
                subdivisions=control.subdivisions,
                localPosition=control.localPosition,
                localRotation=control.localRotation,
                localScale=control.localScale,
                color=control.color,
                opacity=control.opacity,
                lineWidth=control.lineWidth,
                lineStyle=control.lineStyle,
                filled=control.filled,
                xray=control.xray,
            )
        )

    rigDefinition.root = rigRoot.name
    rigDefinition.targetRoot = targetRootPath

    if not rigRoot.hasAttr('presets'):
        savePreset(rigRoot, 'default')

    rigDefinition.loadPresets(rigRoot.presets)

    rigDefinition.save(filePath)


@mx.undoable
def importRig(filePath, targetRoot=None):
    """Import a rig from a definition file."""
    rigDefinition = rigdef.RigDefinition()
    rigDefinition.load(filePath)

    if not targetRoot and rigDefinition.targetRoot:
        try:
            targetRoot = mx.Node(rigDefinition.targetRoot)
        except mx.MayaNodeError:
            raise RiggerError('No target root found.')

    rigRoot = mx.cmd.createNode('transform', name=rigDefinition.root, skipSelect=True)

    # create controls
    for control in rigDefinition.controls:
        name = control.path.split('|')[-1]
        groups = control.path.split('|')[:-1]
        parentPath = f'{rigRoot.pathName}|' + '|'.join(groups)
        target = None

        restPosition = mx.Vector(control.position.x, control.position.y, control.position.z)
        restRotation = mx.Vector(control.rotation.x, control.rotation.y, control.rotation.z)
        restScale = mx.Vector(control.scale.x, control.scale.y, control.scale.z)

        # get control's target
        if control.targetPath is not None:
            try:
                target = mx.Node(f'{targetRoot.pathName}|{ control.targetPath}')
            except mx.MayaNodeError:
                continue

        # create necessary control's groups
        for i in range(0, len(groups)):
            groupPath = f'{rigRoot.pathName}|' + '|'.join(groups[: i + 1])

            if not mx.cmd.objExists(groupPath):
                groupParentPath = f'{rigRoot.pathName}|' + '|'.join(groups[:i])

                mx.cmd.createNode(
                    'transform',
                    name=groups[i],
                    parent=groupParentPath,
                    skipSelect=True,
                )

        placedControl = rigcontrol.createControl(
            target,
            name=name,
            parent=parentPath,
            skipSelect=True,
        )

        placedControl.shapeType = control.shapeType
        placedControl.size = control.size
        placedControl.innerSize = control.innerSize
        placedControl.subdivisions = control.subdivisions
        placedControl.localPosition = mx.Vector(
            control.localPosition.x,
            control.localPosition.y,
            control.localPosition.z,
        )
        placedControl.localRotation = mx.Vector(
            control.localRotation.x,
            control.localRotation.y,
            control.localRotation.z,
        )
        placedControl.localScale = mx.Vector(
            control.localScale.x,
            control.localScale.y,
            control.localScale.z,
        )
        placedControl.color = control.color
        placedControl.opacity = control.opacity
        placedControl.lineWidth = control.lineWidth
        placedControl.lineStyle = control.lineStyle
        placedControl.filled = control.filled
        placedControl.xray = control.xray

        placedControl.restPosition = restPosition
        placedControl.restRotation = restRotation
        placedControl.restScale = restScale

        if not target:
            placedControl.worldPosition = restPosition
            placedControl.worldRotation = restRotation
            placedControl.worldScale = restScale

    presetsAttr = rigRoot.addAttr('presets', rigDefinition.dumpPresets())
    presetsAttr.locked = True

    applyPreset(rigRoot, rigDefinition.presets[0])

    return rigRoot


# --------------------------------------------------------------------------------------------------


@mx.undoable
def applyPreset(rigRoot, preset: rigdef.RigDefinition.Preset):
    """Apply the rig preset."""
    # clear existing IKs and spaces
    for control in preset.controls:
        try:
            placedControl = mx.Node(f'{rigRoot.pathName}|{control.path}')
        except mx.MayaNodeError:
            continue

        ikChain = rigquery.getIKChainFromSingleControl(placedControl)

        if ikChain:
            deleteIKChain(ikChain)

        rigcontrol.changeControlParentSpace(placedControl, None)
        rigcontrol.changeControlPositionSpace(placedControl, None)
        rigcontrol.changeControlRotationSpace(placedControl, None)
        rigcontrol.changeControlScaleSpace(placedControl, None)
        rigcontrol.changeControlAimSpace(placedControl, None)

    # create IK chains
    for ikChain in preset.ikChains:
        try:
            ikControls = [mx.Node(f'{rigRoot.pathName}|{link}') for link in ikChain]
        except mx.MayaNodeError:
            continue

        switchControlsToIK(ikControls)

    # set controls' spaces
    for control in preset.controls:
        try:
            placedControl = mx.Node(f'{rigRoot.pathName}|{control.path}')
        except mx.MayaNodeError:
            continue

        for spaceTypes in [
            [rigcontrol.changeControlParentSpace, control.parentSpaces],
            [rigcontrol.changeControlPositionSpace, control.positionSpaces],
            [rigcontrol.changeControlRotationSpace, control.rotationSpaces],
            [rigcontrol.changeControlScaleSpace, control.scaleSpaces],
            [rigcontrol.changeControlAimSpace, control.aimSpaces],
        ]:
            spaceFn = spaceTypes[0]
            spaces = spaceTypes[1]

            for space in spaces:
                spaceTargetPath = space.target

                if not spaceTargetPath.startswith('|'):
                    spaceTargetPath = f'{rigRoot.pathName}|{spaceTargetPath}'

                try:
                    spaceTarget = mx.Node(spaceTargetPath)
                except mx.MayaNodeError:
                    continue

                spaceFn(
                    placedControl,
                    rigcontrol.SpaceTarget(spaceTarget, space.weight),
                    rigcontrol.SpaceOperation.ADD,
                )

    # TODO: This shouldn't be necessary if `switchControlsToIK` wouldn't select various nodes.
    mx.cmd.select(clear=True)


@mx.undoable
def savePreset(rigRoot, name: str):
    """Save a rig preset."""
    name = sanitizePresetName(name)

    if not name:
        raise RiggerError('Invalid rig preset name.')

    rigRootPath = rigRoot.pathName

    def extractControlSpaces(spaces, ikChain):
        extractedSpaces = []

        for space in spaces:
            if not ikChain or space.target.parent not in ikChain.joints:  # ignore ik offsets
                extractedSpaces.append(
                    rigdef.RigDefinition.PresetSpace(
                        target=space.target.pathName.replace(f'{rigRoot.pathName}|', ''),
                        weight=space.weight,
                    )
                )

        return extractedSpaces

    preset = rigdef.RigDefinition.Preset(name=name)

    for control in rigquery.getAllControls(rigRoot):
        ikChain = rigquery.getIKChainFromSingleControl(control)

        if ikChain:
            ikControls = [
                getObjectPathRelativeToRoot(ikControl, rigRootPath)
                for ikControl in ikChain.controls
            ]

            if ikControls not in preset.ikChains:
                preset.ikChains.append(ikControls)

        preset.controls.append(
            rigdef.RigDefinition.PresetControl(
                path=getObjectPathRelativeToRoot(control, rigRootPath),
                parentSpaces=extractControlSpaces(
                    rigcontrol.getControlParentSpace(control),
                    ikChain,
                ),
                positionSpaces=extractControlSpaces(
                    rigcontrol.getControlPositionSpace(control),
                    ikChain,
                ),
                rotationSpaces=extractControlSpaces(
                    rigcontrol.getControlRotationSpace(control),
                    ikChain,
                ),
                scaleSpaces=extractControlSpaces(
                    rigcontrol.getControlScaleSpace(control),
                    ikChain,
                ),
                aimSpaces=extractControlSpaces(
                    rigcontrol.getControlAimSpace(control),
                    ikChain,
                ),
            )
        )

    if rigRoot.hasAttr('presets'):
        presets = rigdef.RigDefinition().loadPresets(rigRoot.presets).presets
    else:
        presets = []
        rigRoot.addAttr('presets', '')

    for i, existingPreset in enumerate(presets):
        if existingPreset.name == preset.name:
            presets[i] = preset
            break
    else:
        presets.append(preset)

    rigRoot['presets'].locked = False
    rigRoot.presets = rigdef.RigDefinition(presets=presets).dumpPresets()
    rigRoot['presets'].locked = True


@mx.undoable
def deletePreset(rigRoot, name: str):
    """Delete a rig preset."""
    presets = rigquery.getRigPresets(rigRoot)

    for preset in presets:
        if preset.name == name:
            presets.remove(preset)

            rigRoot['presets'].locked = False
            rigRoot.presets = rigdef.RigDefinition(presets=presets).dumpPresets()
            rigRoot['presets'].locked = True

            return


# --------------------------------------------------------------------------------------------------


@mx.undoable
def switchControlsToFK(controls):
    """Switch controls to FK."""
    if not controls:
        return

    for control in controls:
        ikChain = rigquery.getIKChainFromSingleControl(control)

        if ikChain:
            deleteIKChain(ikChain)

            if len(controls) == 1:
                controls = ikChain.controls

    for i in range(len(controls) - 1, 0, -1):
        rigcontrol.changeControlParentSpace(controls[i], controls[i - 1])

    mx.cmd.select(controls)


@mx.undoable
def switchControlsToReversedFK(controls):
    """Switch controls to Reversed FK."""
    if not controls:
        return

    for control in controls:
        ikChain = rigquery.getIKChainFromSingleControl(control)

        if ikChain:
            deleteIKChain(ikChain)

            if len(controls) == 1:
                controls = ikChain.controls

    if len(controls) == 1:
        controls = rigquery.getFKChainFromSingleControl(controls[0])

    for i in range(0, len(controls) - 1):
        rigcontrol.changeControlParentSpace(controls[i], controls[i + 1])

    # select the controls in reversed order to continuously reverse them
    mx.cmd.select(controls[::-1])


@mx.undoable
def switchControlsToIK(controls):
    """Switch controls to IK."""
    if len(controls) == 1:
        controls = rigquery.getFKChainFromSingleControl(controls[0])

    if len(controls) < 3:
        return None

    rootControl = controls[0]
    handleControl = controls[-1]
    jointsChain = []
    jointsOffsets = []

    rootControlParentSpace = rigcontrol.getControlParentSpace(rootControl)
    handleControlParentSpace = rigcontrol.getControlParentSpace(handleControl)

    group = mx.cmd.createNode(
        'transform',
        name='ikChain_grp',
        parent=handleControl,
        skipSelect=True,
    )
    group.inheritsTransform = False
    group.hiddenInOutliner = True
    group.visibility = False

    mx.cmd.select(clear=True)

    for control in controls:
        joint = mx.cmd.joint(position=control.worldPosition, name=f'{control.name}_ikJoint')
        joint.rotateOrder = control.rotateOrder

        if jointsChain:
            mx.cmd.joint(jointsChain[-1], edit=True, orientJoint='xyz')

        joint.addAttr('ikControl', control)

        jointsChain.append(joint)

        # Disconnect spaces to avoid evaluation cycles.
        if control in rootControlParentSpace:
            rigcontrol.changeControlParentSpace(
                rootControl,
                control,
                rigcontrol.SpaceOperation.REMOVE,
            )
        if control in handleControlParentSpace:
            rigcontrol.changeControlParentSpace(
                handleControl,
                control,
                rigcontrol.SpaceOperation.REMOVE,
            )

    jointsChain[0].parent = group

    for index in range(0, len(controls) - 1):
        jointOffset = mx.cmd.createNode(
            'transform',
            name=f'{controls[index].name}_ikOffset',
            parent=jointsChain[index],
            skipSelect=True,
        )
        jointsChain[index].addAttr('offset', jointOffset)
        mx.cmd.matchTransform(jointOffset, controls[index], position=True, rotation=True)
        jointsOffsets.append(jointOffset)

    ikHandle, ikEffector = mx.cmd.ikHandle(startJoint=jointsChain[0], endEffector=jointsChain[-1])
    ikHandle.parent = group
    ikHandle.name = 'ikChain_handle'
    ikEffector.name = 'ikChain_effector'

    mx.cmd.matchTransform(ikHandle, ikEffector, rotation=True)

    for index in range(1, len(controls) - 1):
        rigcontrol.changeControlParentSpace(controls[index], jointsOffsets[index])
        rigcontrol.unlinkControl(controls[index], jointsOffsets[index])

    poleControl = rigcontrol.createPoleVectorControl(
        ikHandle,
        name=f'{controls[1].name}_ikPole',
        parent=handleControl.parent,
        skipSelect=True,
    )
    polePosition, poleRotation = rigutils.calculatePoleVectorPositionAndRotation(
        [
            controls[0].restPosition,
            controls[1].restPosition,
            controls[-1].restPosition,
        ]
    )
    poleControl.restPosition = polePosition
    poleControl.restRotation = poleRotation

    jointsChain[-1].addAttr('ikPoleControl', poleControl)

    mx.cmd.reorder(poleControl, front=True)
    if rootControl.parent:
        mx.cmd.reorder(poleControl, relative=rootControl.parent.children.index(rootControl) + 1)
    else:
        mx.cmd.reorder(poleControl, relative=mx.cmd.ls(assemblies=True).index(rootControl) + 1)

    mx.cmd.pointConstraint(rootControl, jointsChain[0])
    rigcontrol.changeControlRotationSpace(rootControl, jointsOffsets[0])

    ikHandleMatrix = ikHandle.worldMatrix
    handleControl['worldMatrix'].connect(ikHandle['offsetParentMatrix'])
    ikHandle.worldMatrix = ikHandleMatrix

    # add control links
    for i in range(0, len(controls) - 1):
        rigcontrol.linkControl(
            controls[i],
            rigcontrol.LinkTarget(
                controls[i + 1],
                color=controls[i].color,
            ),
        )

    # add pole control link
    rigcontrol.linkControl(
        poleControl,
        rigcontrol.LinkTarget(
            controls[1],
            style=rigcontrol.LinkStyle.DASHED,
            color=poleControl.color,
        ),
    )

    handleControl.select()

    return rigquery.IKChain(
        controls=controls,
        poleControl=poleControl,
        joints=jointsChain,
    )


@mx.undoable
def deleteIKChain(ikChain):
    """Delete IK chain."""
    for i in range(1, len(ikChain.controls) - 1):
        # Clear the parent space of the middle controls
        # to not have them change position when deleting the IK joints.
        rigcontrol.changeControlParentSpace(ikChain.controls[i], None)

    rigcontrol.changeControlRotationSpace(
        ikChain.controls[0],
        ikChain.joints[0].offset,
        rigcontrol.SpaceOperation.REMOVE,
    )

    ikChain.joints[0].parent.delete()

    rigcontrol.deleteControl(ikChain.poleControl)

    # remove control links
    for i in range(0, len(ikChain.controls) - 1):
        rigcontrol.unlinkControl(ikChain.controls[i], ikChain.controls[i + 1])


# --------------------------------------------------------------------------------------------------


@mx.undoable
def moveControlsToRestPose(controls, spreadToRig=False):
    """Move controls to rest pose."""
    if not controls:
        return

    if spreadToRig:
        rigRoots = []
        rigControls = []

        for control in controls:
            rigRoot = rigquery.getRigRootFromControl(control)

            if not rigRoot:
                rigControls.append(control)
            elif rigRoot not in rigRoots:
                rigRoots.append(rigRoot)
                rigControls.extend(rigquery.getAllControls(rigRoot))

        controls = rigControls

    tempNodes = []
    intermediaryNodes = []

    translateAnimationCurves = []
    rotateAnimationCurves = []
    scaleAnimationCurves = []

    def disconnectAnimationCurves(obj, attr):
        curves = [
            obj[f'{attr}X'].input(),
            obj[f'{attr}Y'].input(),
            obj[f'{attr}Z'].input(),
        ]

        if curves[0]:
            obj[f'{attr}X'].disconnectInput()

        if curves[1]:
            obj[f'{attr}Y'].disconnectInput()

        if curves[2]:
            obj[f'{attr}Z'].disconnectInput()

        return curves

    def reconnectAnimationCurves(obj, attr, curves):
        if curves[0]:
            curves[0]['output'].connect(obj[f'{attr}X'])

        if curves[1]:
            curves[1]['output'].connect(obj[f'{attr}Y'])

        if curves[2]:
            curves[2]['output'].connect(obj[f'{attr}Z'])

    for control in controls:
        restNode = mx.cmd.createNode(
            'transform',
            name=f'{control.name}_temp',
            parent=control.parent,
            skipSelect=True,
        )
        restNode.inheritsTransform = False

        restNode.worldPosition = control.restPosition
        restNode.worldRotation = control.restRotation
        restNode.worldScale = control.restScale

        translateAnimationCurves.append(disconnectAnimationCurves(control, 'translate'))
        rotateAnimationCurves.append(disconnectAnimationCurves(control, 'rotate'))
        scaleAnimationCurves.append(disconnectAnimationCurves(control, 'scale'))

        parentConstraint = mx.cmd.parentConstraint(restNode, control)[0]
        scaleConstraint = mx.cmd.scaleConstraint(restNode, control)[0]

        # TODO: Remove if the rig control is no longer using the joint for its transform node.
        scaleConstraint['constraintScaleCompensate'].disconnectInput()
        scaleConstraint.constraintScaleCompensate = False

        scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
        scaleConstraint['constraintScale'].connect(scaleOffsetNode['input1'])
        control['inverseScale'].connect(scaleOffsetNode['input2'])

        scaleOffsetNode['outputX'].connect(control['scaleX'], force=True)
        scaleOffsetNode['outputY'].connect(control['scaleY'], force=True)
        scaleOffsetNode['outputZ'].connect(control['scaleZ'], force=True)

        tempNodes.append(restNode)
        tempNodes.append(parentConstraint)
        tempNodes.append(scaleConstraint)
        intermediaryNodes.append(scaleOffsetNode)

    for node in intermediaryNodes:
        # Needed. Just deleting the `multiplyDivide` nodes changes the transforms.
        node['outputX'].disconnectOutput()
        node['outputY'].disconnectOutput()
        node['outputZ'].disconnectOutput()

    if tempNodes:
        mx.cmd.delete(tempNodes)

    if intermediaryNodes:
        mx.cmd.delete(intermediaryNodes)

    for index, control in enumerate(controls):
        reconnectAnimationCurves(control, 'translate', translateAnimationCurves[index])
        reconnectAnimationCurves(control, 'rotate', rotateAnimationCurves[index])
        reconnectAnimationCurves(control, 'scale', scaleAnimationCurves[index])

        control.worldPosition = control.restPosition
        control.worldRotation = control.restRotation
        control.worldScale = control.restScale

    mx.cmd.select(controls)


@mx.undoable
def deleteControls(controls):
    """Delete the rig controls."""
    for control in controls:
        if control.exists:
            ikChain = rigquery.getIKChainFromSingleControl(control)

            if ikChain:
                deleteIKChain(ikChain)

    for control in controls:
        if control.exists:
            rigcontrol.deleteControl(control)


# --------------------------------------------------------------------------------------------------


def getObjectPathRelativeToRoot(obj, rootPath: str) -> str:
    """Get object's path name relative to a root path."""
    return obj.pathName[len(rootPath) + 1 :]


def sanitizePresetName(name: str) -> str:
    """Sanitize the rig preset name."""
    name = '|'.join(category.strip() for category in name.split('|') if category.strip())
    name = re.sub(r'\s+', ' ', name)

    return name
