# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Settings functionality."""

import os
import maya.cmds as cmds
import puppeteer.ui as ui
import puppeteer.rigcontrol as rigcontrol


SETTINGS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), '..', '..', 'settings.json'
)


class Settings(ui.BindableData):
    """The settings data."""

    class RigControl(ui.BindableData):
        shapeType = rigcontrol.ShapeType.SPHERE
        size = 1.0
        innerSize = 0.0
        subdivisions = 0
        localPosition = (0.0, 0.0, 0.0)
        localRotation = (0.0, 0.0, 0.0)
        localScale = (1.0, 1.0, 1.0)
        color = (1.222, 0.658, 0.064)
        opacity = 1.0
        lineWidth = 1.0
        lineStyle = rigcontrol.LineStyle.SOLID
        filled = True
        xray = True

    class Dev(ui.BindableData):
        hotReloadEnabled = False
        hotReloadExtraPath = ''
        hotReloadExtraPackage = ''

    rigControl = RigControl()
    dev = Dev()


def get(filePath: str = SETTINGS_PATH) -> Settings:
    """Retrieve the settings."""
    if hasattr(get, 'instance'):
        settings = get.instance
    else:
        settings = Settings()

        try:
            with open(filePath, 'r') as dataFile:
                settings.jsonLoad(dataFile.read())
        except FileNotFoundError:
            pass

        get.instance = settings

    return settings


def save(settings: Settings = None, filePath: str = SETTINGS_PATH):
    """Save the settings."""
    if settings:
        get.instance = settings
    else:
        settings = get(filePath)

    with open(filePath, 'w') as dataFile:
        dataFile.write(settings.jsonDump(beautify=True))

    updatePluginSettings(settings)


def clearCache():
    """Clear the cached settings."""
    if hasattr(get, 'instance'):
        del get.instance


def updatePluginSettings(settings: Settings):
    """Update the plugin settings."""
    cmds.puppeteerUpdateSettings(
        rigControlShapeType=settings.rigControl.shapeType,
        rigControlSize=settings.rigControl.size,
        rigControlInnerSize=settings.rigControl.innerSize,
        rigControlSubdivisions=settings.rigControl.subdivisions,
        rigControlLocalPosition=settings.rigControl.localPosition,
        rigControlLocalRotation=settings.rigControl.localRotation,
        rigControlLocalScale=settings.rigControl.localScale,
        rigControlColor=settings.rigControl.color,
        rigControlOpacity=settings.rigControl.opacity,
        rigControlLineWidth=settings.rigControl.lineWidth,
        rigControlLineStyle=settings.rigControl.lineStyle,
        rigControlFilled=settings.rigControl.filled,
        rigControlXray=settings.rigControl.xray,
    )
