# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""The UI for managing settings."""

import PySide2.QtWidgets as qtw
import PySide2.QtGui as qtg
import puppeteer.mayax as mx
import puppeteer.rigcontrol as rigcontrol
import puppeteer.ui as ui
import puppeteer.settings as settings


class MainWindow(ui.SingletonWindow):
    """Settings main window."""

    def __init__(self):
        super().__init__('Settings - Puppeteer')

        self.setMinimumWidth(500)
        self.setMinimumHeight(600)

        self.settings = settings.get().copy()

        self.createUI()

    def createUI(self):
        """Create the UI."""
        with qtw.QVBoxLayout() as mainLayout:
            self.setLayout(mainLayout)

            mainLayout.setSpacing(0)
            mainLayout.setContentsMargins(10, 10, 10, 10)

            with ui.ScrollVBox():
                createRigControlPropertiesUI(self.settings)

            with qtw.QHBoxLayout() as layout:
                layout.setSpacing(5)

                with qtw.QPushButton('Save') as saveBtn:
                    saveBtn.clicked.connect(self.saveSettings)

                with qtw.QPushButton('Cancel') as cancelBtn:
                    cancelBtn.clicked.connect(self.close)

    def saveSettings(self):
        """Save the settings."""
        settings.save(self.settings)

        self.close()


def createRigControlPropertiesUI(theSettings: settings.Settings):
    """Create the UI for the rig control's default properties."""
    with ui.GroupVBox('Rig Control'):
        with qtw.QFormLayout() as formLayout:
            shapeTypeField = qtw.QComboBox()
            shapeTypeField.setSizePolicy(qtw.QSizePolicy.Expanding, qtw.QSizePolicy.Minimum)

            for shapeType in rigcontrol.ShapeType:
                shapeTypeField.addItem(shapeType.name.lower(), shapeType.value)

            shapeTypeField.insertSeparator(5)
            shapeTypeField.insertSeparator(12)

            sizeValidator = qtg.QDoubleValidator()
            sizeValidator.setBottom(0)
            sizeValidator.setDecimals(3)
            sizeField = qtw.QLineEdit()
            sizeField.setValidator(sizeValidator)

            innerSizeField = ui.Slider(0.0, 1.0)
            subdivisionsField = ui.Slider(0, 50)
            opacityField = ui.Slider(0.0, 1.0)
            lineWidthField = ui.Slider(0.0, 10.0)

            localPositionField = ui.VectorField()
            localRotationField = ui.VectorField()
            localScaleField = ui.VectorField()
            colorField = ui.ColorField(renderingSpace=True)

            lineStyleField = qtw.QComboBox()
            lineStyleField.setSizePolicy(qtw.QSizePolicy.Expanding, qtw.QSizePolicy.Minimum)

            for lineStyle in rigcontrol.LineStyle:
                lineStyleField.addItem(lineStyle.name.lower(), lineStyle.value)

            filledField = qtw.QCheckBox('Filled')
            xrayField = qtw.QCheckBox('X-Ray')

            ui.bind(shapeTypeField, theSettings.rigControl, 'shapeType')
            ui.bind(sizeField, theSettings.rigControl, 'size')
            ui.bind(innerSizeField, theSettings.rigControl, 'innerSize')
            ui.bind(subdivisionsField, theSettings.rigControl, 'subdivisions')
            ui.bind(localPositionField, theSettings.rigControl, 'localPosition')
            ui.bind(localRotationField, theSettings.rigControl, 'localRotation')
            ui.bind(localScaleField, theSettings.rigControl, 'localScale')
            ui.bind(colorField, theSettings.rigControl, 'color')
            ui.bind(opacityField, theSettings.rigControl, 'opacity')
            ui.bind(lineWidthField, theSettings.rigControl, 'lineWidth')
            ui.bind(lineStyleField, theSettings.rigControl, 'lineStyle')
            ui.bind(filledField, theSettings.rigControl, 'filled')
            ui.bind(xrayField, theSettings.rigControl, 'xray')

            formLayout.addRow('Shape', shapeTypeField)
            formLayout.addRow('Size', sizeField)
            formLayout.addRow('Inner Size', innerSizeField)
            formLayout.addRow('Subdivisions', subdivisionsField)

            formLayout.addRow('', ui.LayoutSeparator())

            formLayout.addRow('Position', localPositionField)
            formLayout.addRow('Rotation', localRotationField)
            formLayout.addRow('Scale', localScaleField)

            formLayout.addRow('', ui.LayoutSeparator())

            formLayout.addRow('Color', colorField)
            formLayout.addRow('Opacity', opacityField)
            formLayout.addRow('Line Width', lineWidthField)
            formLayout.addRow('Line Style', lineStyleField)
            formLayout.addRow('', filledField)  # empty label to use CheckBox's text
            formLayout.addRow('', xrayField)  # empty label to use CheckBox's text
