# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Setup functionality."""

from functools import partial
import os
import sys
import webbrowser

import maya.cmds as cmds

import puppeteer.mayax as mx
import puppeteer.utils as utils
import puppeteer.settings as settings
import puppeteer.ui as ui
import puppeteer.log as log


MODULE_FILENAME = 'puppeteer.mod'
PLUGIN_NAME = 'puppeteer'


# --------------------------------------------------------------------------------------------------


def getRootDirectory():
    """Get the root directory."""
    return os.path.normpath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '..',
            '..',
        )
    )


def getSourceDirectory():
    """Get the source directory."""
    return os.path.join(getRootDirectory(), 'src')


def getPluginsDirectory():
    """Get the plugins directory."""
    return os.path.join(getRootDirectory(), 'plugins')


def getModuleFilename():
    """Get the module filename."""
    return os.path.join(mx.getModulesDirectory(), MODULE_FILENAME)


def getPluginFilename():
    """Get the plugin filename."""
    return os.path.join(getPluginsDirectory(), PLUGIN_NAME)


# --------------------------------------------------------------------------------------------------


def install():
    """Install puppeteer."""
    mayaModulesPath = mx.getModulesDirectory()

    if not os.path.exists(mayaModulesPath):
        os.makedirs(mayaModulesPath)

    with open(getModuleFilename(), mode='w+t') as moduleFile:
        moduleContent = (
            f'+ Puppeteer 1.0 {getRootDirectory()}\n'
            f'scripts: {getSourceDirectory()}\n'
            f'plug-ins: {getPluginsDirectory()}'
        )

        moduleFile.write(moduleContent)

    pluginFilename = getPluginFilename()
    cmds.loadPlugin(pluginFilename)
    cmds.pluginInfo(pluginFilename, edit=True, autoload=True)

    createMainMenu()

    cmds.confirmDialog(
        title='Puppeteer',
        message='Puppeteer installed.',
        button='OK',
        icon='information',
    )


def uninstall():
    """Remove puppeteer."""
    sourceDir = getSourceDirectory()
    moduleFilename = getModuleFilename()

    for window in ui.findWidgetsByClass(ui.Window):
        window.close()

    if os.path.exists(moduleFilename):
        os.remove(moduleFilename)

    if sourceDir in sys.path:
        sys.path.remove(sourceDir)

    cmds.unloadPlugin(PLUGIN_NAME, force=True)
    cmds.pluginInfo(PLUGIN_NAME, edit=True, autoload=False)

    removeMainMenu()

    log.info('Uninstalled.')

    utils.removePyModules('puppeteer')


# --------------------------------------------------------------------------------------------------


def createMainMenu():
    """Add puppeteer menu to Maya's main menu bar."""
    hotReloadEnabled = settings.get().dev.hotReloadEnabled

    cmds.menu(
        'puppeteerMainMenu',
        parent='MayaWindow',
        label='Puppeteer',
        tearOff=True,
    )

    cmds.menuItem(divider=True, label="Tools")
    # cmds.menuItem(label='Rigger', command=onRiggerMenu)
    cmds.menuItem(label='Animator', command=onAnimatorMenu)

    cmds.menuItem(divider=True, label="Preferences")
    cmds.menuItem(label='Settings', command=onSettingsMenu)

    cmds.menuItem(divider=True, label="Help")
    cmds.menuItem(label='Docs (online)', command=onDocsMenu)
    cmds.menuItem(label='Docs (offline)', command=partial(onDocsMenu, offline=True))

    cmds.menuItem(divider=True, label="Developer")
    cmds.menuItem(
        'puppeteerHotReloadMenuItem',
        label='Hot Reload',
        checkBox=hotReloadEnabled,
        command=onHotReloadMenu,
    )

    cmds.menuItem(divider=True)
    cmds.menuItem(label='Uninstall', command=onUninstallMenu)

    if hotReloadEnabled:
        onHotReloadMenu(True)


def removeMainMenu():
    """Remove puppeteer menu."""
    cmds.deleteUI('puppeteerMainMenu')


# --------------------------------------------------------------------------------------------------
# Menu Actions
# --------------------------------------------------------------------------------------------------


def onRiggerMenu(*_):
    """Show the Rigger main window."""
    from puppeteer.rigger_ui import MainWindow

    MainWindow().show()


def onAnimatorMenu(*_):
    """Show the Animator main window."""
    from puppeteer.animator_ui import MainWindow

    MainWindow().show()


def onSettingsMenu(*_):
    """Show the settings window."""
    from puppeteer.settings_ui import MainWindow

    MainWindow().show()


def onUninstallMenu(*_):
    """Call uninstall."""
    confirm = cmds.confirmDialog(
        title='Puppeteer',
        message='Are you sure you want to uninstall Puppeteer?',
        button=('Yes', 'No'),
        defaultButton='Yes',
        cancelButton='No',
        icon='information',
    )

    if confirm == 'Yes':
        uninstall()


def onDocsMenu(*_, offline=False):
    """Open the documentation."""
    docsUrl = 'https://puppeteer.movindot.com/docs/'

    if offline:
        root = getRootDirectory().replace('\\', '/')
        docsUrl = f'file:///{root}/docs/index.html'

    webbrowser.open(docsUrl)


def onHotReloadMenu(enable):
    """Enable/disable Hot Reload."""
    from puppeteer import hotreload

    if enable:
        hotreload.enable()
    else:
        hotreload.disable()

    cmds.menuItem('puppeteerHotReloadMenuItem', edit=True, label='Hot Reload')

    settings.get().dev.hotReloadEnabled = enable
    settings.save()
