# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Data binding functionality."""

import copy
import json
import PySide2.QtWidgets as qtw
import PySide2.QtGui as qtg
import maya.api.OpenMaya as om
import puppeteer.mayax as mx
import puppeteer.utils as utils
from .widgets import (
    GroupBox,
    NodeField,
    ColorField,
    Slider,
    ToggleButton,
    VectorField,
)


class UIDataBindingError(Exception):
    """Raised by UI's data binding."""


class BindingType:
    """Binding type."""

    VALUE = 1
    VISIBILITY = 2
    INVISIBILITY = 3
    CREATION = 4


class BindableDataEncoder(json.JSONEncoder):
    """Helper for JSON encoding."""

    def __init__(self, objEncoder, **kwargs):
        super().__init__(**kwargs)

        self.objEncoder = objEncoder

    def default(self, o):  # pylint: disable=method-hidden
        """Encode an object."""
        if isinstance(o, BindableData):
            return {key: value for key, value in o.__dict__.items() if not key.startswith('_')}
        elif isinstance(o, om.MVector):
            return {'x': o.x, 'y': o.y, 'z': o.z}
        elif self.objEncoder:
            try:
                return self.objEncoder(o)
            except TypeError:
                pass

        return super().default(o)


class BindableData:
    """An object that can be used for data binding."""

    # Use __slots__ to not pollute __dict__ with internal attributes.
    __slots__ = ['__boundProperties__', '__dict__']

    def __init__(self, **properties):
        super().__setattr__('__boundProperties__', {})

        defaultProperties = self.__getDefaultProperties()

        for name in defaultProperties:
            self.__dict__[name] = _validateBoundValue(defaultProperties[name])

        for name in properties:
            if name not in self.__dict__ and self.__class__ is not BindableData:
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{name}'"
                )

            self.__dict__[name] = _validateBoundValue(properties[name])

    @classmethod
    def __getDefaultProperties(cls):
        properties = {}
        classProperties = cls.__dict__

        for name in classProperties:
            isValidProperty = (
                not name.startswith('_')
                and not callable(classProperties[name])
                and not isinstance(classProperties[name], property)
                and not isinstance(classProperties[name], classmethod)
                and not isinstance(classProperties[name], staticmethod)
            )

            if isValidProperty:
                if isinstance(classProperties[name], BindableData):
                    properties[name] = copy.deepcopy(classProperties[name])
                else:
                    properties[name] = classProperties[name]

        return properties

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
            return

        try:
            boundProperty = self.__dict__[name]
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

        if isinstance(boundProperty, BindableData):
            for propertyName in value:
                boundProperty[propertyName] = value[propertyName]
        elif isinstance(boundProperty, BindableList):
            del boundProperty[:]
            boundProperty.extend(value)
        elif isinstance(boundProperty, BindableNode):
            boundProperty.__boundNode__ = value
        else:
            super().__setattr__(name, value)

            _syncBoundProperties(self, name, value)

    def __getitem__(self, name):
        return getattr(self, name)

    def __setitem__(self, name, value):
        self.__setattr__(name, value)

    def __iter__(self):
        return self.__dict__.__iter__()

    def __repr__(self):
        properties = ', '.join([propertyName for propertyName in self])

        return f'{self.__class__.__name__}({properties})'

    def __str__(self):
        return str(self.__dict__)

    def clear(self):
        """Clear all data."""
        defaultProperties = self.__getDefaultProperties()

        for propertyName in self.__dict__:
            propertyValue = self[propertyName]

            if isinstance(propertyValue, BindableData):
                propertyValue.clear()
            elif isinstance(propertyValue, BindableList):
                del propertyValue[:]
            else:
                if propertyName in defaultProperties:
                    defaultValue = defaultProperties[propertyName]
                elif isinstance(propertyValue, (bool, int, float, str)):
                    defaultValue = type(propertyValue)()
                else:
                    defaultValue = None

                self.__setattr__(propertyName, defaultValue)

    def jsonDump(self, beautify=False, objEncoder=None):
        """Encode to JSON.

        Parameters
        ----------
        beautify : bool, optional
            If `True` the encoded string will be more human-friendly.
        objEncoder : callable, optional
            Function called to encode an object. Should raise TypeError for default implementation.

        Returns
        -------
        str
            The encoded JSON string.
        """
        return json.dumps(
            self,
            cls=BindableDataEncoder,
            objEncoder=objEncoder,
            sort_keys=False,
            indent=4 if beautify else None,
            separators=(',', ': ') if beautify else (',', ':'),
        )

    def jsonLoad(self, jsonData):
        """Decode from JSON."""
        data = json.loads(jsonData)

        for key in data:
            self.__setattr__(key, data[key])

        return self

    def copy(self):
        """Retrieve a copy."""
        settings = self.__class__()
        settings.jsonLoad(self.jsonDump())

        return settings


class BindableNode:
    """Wraps a `Node` object in order to allow data binding."""

    __slots__ = [
        '__boundProperties__',
        '__boundNode__',
        '__boundNodeAttrChangedCallbackId__',
        '__boundSyncEnabled__',
        '__weakref__',
    ]

    def __init__(self, node):
        self.__boundProperties__ = {}
        self.__boundNodeAttrChangedCallbackId__ = 0
        self.__boundNode__ = node
        self.__boundSyncEnabled__ = True

    def __del__(self):
        if self.__boundNodeAttrChangedCallbackId__:
            om.MNodeMessage.removeCallback(self.__boundNodeAttrChangedCallbackId__)

    def __getattr__(self, name):
        return getattr(self.__boundNode__, name)

    def __setattr__(self, name, value):
        if name not in self.__slots__:
            setattr(self.__boundNode__, name, value)
        else:
            super().__setattr__(name, value)

            if name == '__boundNode__':
                if self.__boundNodeAttrChangedCallbackId__:
                    om.MNodeMessage.removeCallback(self.__boundNodeAttrChangedCallbackId__)
                    self.__boundNodeAttrChangedCallbackId__ = 0

                if value:
                    callbackId = om.MNodeMessage.addAttributeChangedCallback(
                        value.apiObject,
                        utils.weakMethodProxy(self._onNodeAttrChanged),
                    )
                    self.__boundNodeAttrChangedCallbackId__ = callbackId

                    for propertyName in self.__boundProperties__:
                        _syncBoundProperties(
                            self,
                            propertyName,
                            getattr(self.__boundNode__, propertyName),
                        )

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __setitem__(self, name, value):
        self.__setattr__(name, value)

    def _onNodeAttrChanged(self, attrMsg, plug, *_):
        if not self.__boundSyncEnabled__ or not attrMsg & om.MNodeMessage.kAttributeSet:
            return

        attrName = plug.partialName(useLongNames=True)

        _syncBoundProperties(self, attrName, getattr(self.__boundNode__, attrName))


class BindableList(list):
    """A list that can be used for data binding."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__boundSequence__ = []

        for i in range(0, len(self)):
            super().__setitem__(i, _validateBoundValue(self[i]))

    def __iadd__(self, iterable):
        previousLength = len(self)

        super().__iadd__(iterable)

        for i in range(previousLength, len(self)):
            super().__setitem__(i, _validateBoundValue(self[i]))

        _syncBoundSequence(self)

        return self

    def __imul__(self, scalar):
        super().__imul__(scalar)

        _syncBoundSequence(self)

        return self

    def __setitem__(self, index, item):
        if isinstance(index, slice):
            super().__setitem__(index, item)

            for i in range(0, len(self)):
                super().__setitem__(i, _validateBoundValue(self[i]))

            _syncBoundSequence(self)
        else:
            item = _validateBoundValue(item)

            super().__setitem__(index, item)

            _syncBoundSequenceUpdate(self, index, item)

    def __delitem__(self, index):
        super().__delitem__(index)

        if isinstance(index, slice):
            _syncBoundSequence(self)
        else:
            _syncBoundSequenceRemoval(self, index)

    def append(self, item):
        """Add an item to the end of the list."""
        item = _validateBoundValue(item)

        super().append(item)

        _syncBoundSequenceInsertion(self, len(self) - 1, item)

    def extend(self, iterable):
        """Extend the list by appending all the items from the iterable."""
        previousLength = len(self)

        super().extend(iterable)

        for i in range(previousLength, len(self)):
            super().__setitem__(i, _validateBoundValue(self[i]))

        _syncBoundSequence(self)

    def insert(self, index, item):
        """Insert an item at a given position."""
        item = _validateBoundValue(item)

        super().insert(index, item)

        insertionIndex = index

        if index >= len(self):
            # Append behaviour if index is out of range.
            insertionIndex = len(self) - 1

        _syncBoundSequenceInsertion(self, insertionIndex, item)

    def pop(self, index):
        """Remove the item at the given position in the list, and return it."""
        item = super().pop(index)

        _syncBoundSequenceRemoval(self, index)

        return item

    def remove(self, item):
        """Remove the first found item from the list."""
        try:
            index = self.index(item)

            super().remove(item)

            _syncBoundSequenceRemoval(self, index)
        except ValueError:
            raise

    def reverse(self):
        """Reverse the elements of the list in place."""
        super().reverse()

        _syncBoundSequence(self)

    def sort(self, *args, **kwargs):
        """Sort the items of the list in place."""
        super().sort(*args, **kwargs)

        _syncBoundSequence(self)


class BindableIndex:
    """A custom integer that can be used as a mutable/bindable index."""

    def __init__(self, value):
        self.__boundProperties__ = {}
        self.value = int(value)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        _syncBoundProperties(self, name, value)

    def __int__(self):
        return self.value

    def __bool__(self):
        return bool(self.value)

    def __index__(self):
        return self.value

    def __hash__(self):
        return hash(self.value)

    def __nonzero__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other

    def __add__(self, value):
        return self.value + value

    def __sub__(self, value):
        return self.value - value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'BindableIndex({self.value})'


# --------------------------------------------------------------------------------------------------


def bind(widget, obj, propertyName, valueProcessor=None):
    """Bind to the widget's data value."""
    obj = _validateBoundObject(obj)

    if isinstance(obj, list) and callable(propertyName):
        _bindSequence(widget, obj, propertyName)
    else:
        _bindProperty(widget, obj, propertyName, BindingType.VALUE, valueProcessor)


def bindVisibility(widget, obj, propertyName):
    """Bind to the widget's visibility."""
    obj = _validateBoundObject(obj)

    _bindProperty(widget, obj, propertyName, BindingType.VISIBILITY)


def bindInvisibility(widget, obj, propertyName):
    """Bind to the widget's invisibility."""
    obj = _validateBoundObject(obj)

    _bindProperty(widget, obj, propertyName, BindingType.INVISIBILITY)


def bindCreation(container, obj, propertyName, creator):
    """Bind widget's creation."""
    _bindProperty(container, obj, propertyName, BindingType.CREATION, creator)


# --------------------------------------------------------------------------------------------------


def _bindProperty(widget, obj, propertyName, bindingType, valueProcessor=None):
    boundProps = obj.__boundProperties__

    if bindingType == BindingType.CREATION:
        bindingData = {
            'widget': None,
            'containerView': widget,
            'bindingType': bindingType,
            'viewCreator': valueProcessor,
        }
    else:
        bindingData = {
            'widget': widget,
            'bindingType': bindingType,
            'valueProcessor': valueProcessor,
        }

    if propertyName in boundProps:
        boundProps[propertyName].append(bindingData)
    else:
        boundProps[propertyName] = [bindingData]

    widget.destroyed.connect(lambda: boundProps[propertyName].remove(bindingData))

    propertyValue = obj[propertyName]

    if bindingType == BindingType.VALUE:
        if isinstance(widget, NodeField):
            widget.nodePicked.connect(
                lambda nodeName: _onWidgetValueChanged(widget, obj, propertyName, nodeName)
            )
        elif isinstance(widget, ColorField):
            widget.colorPicked.connect(
                lambda color: _onWidgetValueChanged(widget, obj, propertyName, color)
            )
        elif isinstance(widget, VectorField):
            widget.valueChanged.connect(
                lambda value: _onWidgetValueChanged(widget, obj, propertyName, value)
            )
        elif isinstance(widget, ToggleButton):
            widget.stateChanged.connect(
                lambda state: _onWidgetValueChanged(widget, obj, propertyName, state)
            )
        elif isinstance(widget, Slider):
            widget.valueChanged.connect(
                lambda value: _onWidgetValueChanged(widget, obj, propertyName, value)
            )
        elif isinstance(widget, GroupBox):
            widget.collapseToggled.connect(
                lambda collapsed: _onWidgetValueChanged(widget, obj, propertyName, collapsed)
            )
        elif isinstance(widget, qtw.QLineEdit):
            widget.textEdited.connect(
                lambda text: _onWidgetValueChanged(widget, obj, propertyName, text)
            )
        elif isinstance(widget, qtw.QDoubleSpinBox):
            widget.valueChanged.connect(
                lambda value: _onWidgetValueChanged(widget, obj, propertyName, value)
            )
        elif isinstance(widget, qtw.QComboBox):
            widget.currentIndexChanged.connect(
                lambda index: _onWidgetValueChanged(widget, obj, propertyName, index)
            )
        elif isinstance(widget, qtw.QCheckBox):
            widget.stateChanged.connect(
                lambda state: _onWidgetValueChanged(widget, obj, propertyName, bool(state))
            )
        elif isinstance(widget, (qtw.QLabel, qtw.QPushButton)):
            # one-way binding
            pass
        else:
            widgetClassName = widget.__class__.__name__
            raise UIDataBindingError(
                f'No "value changed" implementation for widget "{widgetClassName}".'
            )

        _updateWidgetValue(widget, propertyValue, valueProcessor)
    elif bindingType == BindingType.VISIBILITY:
        widget.setVisible(bool(propertyValue))
    elif bindingType == BindingType.INVISIBILITY:
        widget.setHidden(bool(propertyValue))
    elif bindingType == BindingType.CREATION and propertyValue:
        _syncBoundProperties(obj, propertyName, propertyValue)


def _bindSequence(containerWidget, obj, itemViewCreator):
    bindingData = {
        'containerView': containerWidget,
        'itemViewCreator': itemViewCreator,
        'itemsViews': [],
    }

    obj.__boundSequence__.append(bindingData)

    containerWidget.destroyed.connect(lambda: obj.__boundSequence__.remove(bindingData))

    _syncBoundSequence(obj, len(obj.__boundSequence__) - 1)


# --------------------------------------------------------------------------------------------------


def _syncBoundProperties(obj, propertyName, value, widgetInstigator=None):
    boundProps = obj.__boundProperties__

    try:
        boundWidgets = boundProps[propertyName]
    except KeyError:
        return

    for boundData in boundWidgets:
        widget, bindingType = boundData['widget'], boundData['bindingType']

        if widgetInstigator and widgetInstigator is widget:
            continue

        if bindingType == BindingType.VALUE:
            _updateWidgetValue(widget, value, boundData['valueProcessor'])
        elif bindingType == BindingType.VISIBILITY:
            widget.setVisible(bool(value))
        elif bindingType == BindingType.INVISIBILITY:
            widget.setHidden(bool(value))
        elif bindingType == BindingType.CREATION:
            if widget:
                widget.setVisible(False)
                widget.deleteLater()
                boundData['widget'] = None

            if value:
                widget = boundData['viewCreator']()
                boundData['containerView'].addWidget(widget)
                boundData['widget'] = widget


def _syncBoundSequence(obj, widgetIndex=None):
    if widgetIndex:
        boundSequenceData = [obj.__boundSequence__[widgetIndex]]
    else:
        boundSequenceData = obj.__boundSequence__

    for boundData in boundSequenceData:
        itemsViews = boundData['itemsViews']

        if isinstance(boundData['containerView'], qtw.QComboBox):
            boundData['containerView'].clear()
        else:
            for item in itemsViews:
                item['view'].deleteLater()

        del itemsViews[:]

    for index, item in enumerate(obj):
        for boundData in boundSequenceData:
            boundIndex = BindableIndex(index)
            itemView = boundData['itemViewCreator'](item, boundIndex, obj)

            if isinstance(boundData['containerView'], qtw.QComboBox):
                boundData['containerView'].addItem(itemView)
            else:
                boundData['containerView'].addWidget(itemView)

            boundData['itemsViews'].append({'index': boundIndex, 'view': itemView})


def _syncBoundSequenceUpdate(obj, index, item):
    for boundData in obj.__boundSequence__:
        boundView = boundData['itemsViews'][index]
        boundIndex = boundView['index']

        if isinstance(boundData['containerView'], qtw.QComboBox):
            boundView['view'] = boundData['itemViewCreator'](item, boundIndex, obj)
            boundData['containerView'].setItemText(index, boundView['view'])
        else:
            boundView['view'].deleteLater()
            boundView['view'] = boundData['itemViewCreator'](item, boundIndex, obj)

            boundData['containerView'].insertWidget(index, boundView['view'])


def _syncBoundSequenceInsertion(obj, index, item):
    for boundData in obj.__boundSequence__:
        itemsViews = boundData['itemsViews']

        boundIndex = BindableIndex(index)
        itemView = boundData['itemViewCreator'](item, boundIndex, obj)

        if isinstance(boundData['containerView'], qtw.QComboBox):
            boundData['containerView'].insertItem(index, itemView)
        else:
            boundData['containerView'].insertWidget(index, itemView)

        itemsViews.insert(index, {'index': boundIndex, 'view': itemView})

        for i in range(index, len(itemsViews)):
            itemsViews[i]['index'].value = i


def _syncBoundSequenceRemoval(obj, index):
    for boundData in obj.__boundSequence__:
        itemsViews = boundData['itemsViews']

        if isinstance(boundData['containerView'], qtw.QComboBox):
            boundData['containerView'].removeItem(index)
        else:
            itemsViews[index]['view'].deleteLater()

        del itemsViews[index]

        for i in range(index, len(itemsViews)):
            itemsViews[i]['index'].value = i


# --------------------------------------------------------------------------------------------------


def _validateBoundObject(obj):
    if isinstance(obj, mx.Node):
        return BindableNode(obj)

    return obj


def _validateBoundValue(value):
    if isinstance(value, list) and not isinstance(value, BindableList):
        return BindableList(value)
    elif isinstance(value, dict):
        return BindableData(**value)

    return value


def _onWidgetValueChanged(widget, obj, propertyName, value):
    # Update object value without triggering `__setattr__` or `__setitem__`;
    # do that to be able to pass an instigator widget to the sync process.
    try:
        if isinstance(widget, qtw.QLineEdit):
            if not widget.hasAcceptableInput():
                return

            if isinstance(widget.validator(), qtg.QIntValidator):
                value = int(value)
            elif isinstance(widget.validator(), qtg.QDoubleValidator):
                value = float(value)
    except ValueError:
        return

    if isinstance(widget, qtw.QComboBox):
        itemData = widget.itemData(value)

        if itemData is not None:
            value = itemData
        else:
            value = widget.itemText(value)

    if isinstance(obj, list):
        list.__setitem__(obj, propertyName, value)
    if isinstance(obj, BindableNode):
        obj.__boundSyncEnabled__ = False
        obj[propertyName] = value
        obj.__boundSyncEnabled__ = True
    else:
        obj.__dict__[propertyName] = value

    _syncBoundProperties(obj, propertyName, value, widget)


def _updateWidgetValue(widget, value, valueProcessor):
    if valueProcessor:
        value = valueProcessor(value)

    valueText = str(value)

    widget.blockSignals(True)

    if isinstance(widget, NodeField):
        if value:
            widget.setNode(valueText)
        else:
            widget.setNode(None)
    elif isinstance(widget, ColorField):
        widget.setColor(tuple(value))
    elif isinstance(widget, VectorField):
        widget.setValue(tuple(value))
    elif isinstance(widget, ToggleButton):
        widget.setState(value)
    elif isinstance(widget, Slider):
        widget.setValue(value)
    elif isinstance(widget, GroupBox):
        widget.collapse(value)
    elif isinstance(widget, (qtw.QLineEdit, qtw.QLabel, qtw.QPushButton)):
        widget.setText(valueText)
    elif isinstance(widget, qtw.QDoubleSpinBox):
        widget.setValue(value)
    elif isinstance(widget, qtw.QComboBox):
        itemIndex = widget.findData(value)

        if itemIndex > -1:
            widget.setCurrentIndex(itemIndex)
        elif valueText:
            if widget.findText(valueText) == -1:
                raise UIDataBindingError(f'Value "{valueText}" was not found in QComboBox.')

            widget.setCurrentText(valueText)
        else:
            widget.setCurrentIndex(-1)
    elif isinstance(widget, qtw.QCheckBox):
        widget.setChecked(bool(value))
    else:
        widget.blockSignals(False)

        widgetClassName = widget.__class__.__name__
        raise UIDataBindingError(
            f'No "value update" implementation for widget "{widgetClassName}".'
        )

    widget.blockSignals(False)
