# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Custom UI widgets."""

from __future__ import annotations
import typing as t
from functools import partial
import re
import PySide2.QtCore as qt
import PySide2.QtWidgets as qtw
import PySide2.QtGui as qtg
import puppeteer.mayax as mx
import puppeteer.log as log
from .utils import loadIcon


if mx.cmd.optionVar(query='interfaceScalingMode') == 1:  # custom scaling
    SCALE_FACTOR = mx.cmd.optionVar(query='interfaceScalingValue')
else:
    SCALE_FACTOR = 1.0


def addStyleClass(widget, className):
    """Add a style sheet class name to the widget."""
    classProperty = widget.property('class') or ''
    classList = re.split(r'\s+', classProperty.strip())

    for name in re.split(r'\s+', className.strip()):
        if name not in classList:
            classList.append(name)

    widget.setProperty('class', ' '.join(classList))
    widget.setStyleSheet(widget.styleSheet())  # refresh


def removeStyleClass(widget, className):
    """Remove a style sheet class name from the widget."""
    classProperty = widget.property('class') or ''
    classList = re.split(r'\s+', classProperty.strip())

    for name in re.split(r'\s+', className.strip()):
        if name in classList:
            classList.remove(name)

    widget.setProperty('class', ' '.join(classList))
    widget.setStyleSheet(widget.styleSheet())  # refresh


def getStyleClass(widget):
    """Return the style sheet class for the widget."""
    return widget.property('class') or ''


class LayoutSeparator(qtw.QWidget):
    """Create a line to separate the widgets in a layout."""

    def __init__(self):
        super().__init__()

        addStyleClass(self, 'layout-separator')

        self.setAttribute(qt.Qt.WA_StyledBackground)
        self.setSizePolicy(qtw.QSizePolicy.Expanding, qtw.QSizePolicy.Minimum)


class ScrollBox(qtw.QScrollArea):
    """Create a scrollable box area."""

    def __init__(self, boxLayout):
        super().__init__()

        addStyleClass(self, 'scroll-box')

        boxLayout.setContentsMargins(0, 0, 0, 0)

        self._container = qtw.QWidget()
        self._container.setLayout(boxLayout)

        self.setWidget(self._container)
        self.setWidgetResizable(True)
        self.setFocusPolicy(qt.Qt.NoFocus)

    def addWidget(self, widget):
        """Add widget to the end of the layout."""
        self._container.layout().addWidget(widget)

    def insertWidget(self, index, widget):
        """Insert widget at position index."""
        self._container.layout().insertWidget(index, widget)

    def layout(self):
        """Return the used `QLayout`."""
        return self._container.layout()


class ScrollHBox(ScrollBox):
    """Create a scrollable horizontal box area."""

    def __init__(self):
        boxLayout = qtw.QHBoxLayout()
        boxLayout.setAlignment(qt.Qt.AlignLeft)

        super().__init__(boxLayout)

        self.setVerticalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)


class ScrollVBox(ScrollBox):
    """Create a scrollable vertical box area."""

    def __init__(self):
        boxLayout = qtw.QVBoxLayout()
        boxLayout.setAlignment(qt.Qt.AlignTop)

        super().__init__(boxLayout)

        self.setHorizontalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)


# --------------------------------------------------------------------------------------------------


class GroupBox(qtw.QWidget):
    """Create a box area that can be collapsed."""

    collapseToggled = qt.Signal(bool)

    def __init__(self, title='', icon=None, collapsable=True, collapsed=False):
        super().__init__()

        addStyleClass(self, 'group-box')

        self._mainLayout = qtw.QVBoxLayout()
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self._mainLayout.setSpacing(0)

        super().setLayout(self._mainLayout)

        self._titleBar = qtw.QWidget()
        addStyleClass(self._titleBar, 'group-box__title-bar')

        self._contentBox = qtw.QWidget()
        addStyleClass(self._contentBox, 'group-box__content-box')

        self._mainLayout.addWidget(self._titleBar)
        self._mainLayout.addWidget(self._contentBox)

        self._titleBarLabel = qtw.QPushButton(title)
        addStyleClass(self._titleBarLabel, 'group-box__title-label')

        self._buttonsBarLayout = qtw.QHBoxLayout()
        self._buttonsBarLayout.setContentsMargins(4, 0, 0, 0)
        self._buttonsBarLayout.setSpacing(0)
        self._buttonsBar = qtw.QWidget()
        self._buttonsBar.setLayout(self._buttonsBarLayout)
        addStyleClass(self._buttonsBar, 'group-box__buttons-bar')

        titleBarLayout = qtw.QHBoxLayout()
        titleBarLayout.setContentsMargins(0, 0, 0, 0)
        titleBarLayout.setSpacing(0)
        titleBarLayout.addWidget(self._titleBarLabel, 1)
        titleBarLayout.addWidget(self._buttonsBar)
        self._titleBar.setLayout(titleBarLayout)

        if icon:
            iconWidget = qtw.QPushButton()
            iconWidget.setIcon(icon)
            titleBarLayout.insertWidget(0, iconWidget)
            addStyleClass(iconWidget, 'group-box__title-icon')
            addStyleClass(self._titleBar, 'group-box__title-bar--with-icon')

        if collapsable:
            self._collapseTrigger = qtw.QPushButton()

            self._collapseTrigger.clicked.connect(self._onCollapseToggle)
            self._titleBarLabel.clicked.connect(self._onCollapseToggle)
            if icon:
                iconWidget.clicked.connect(self._onCollapseToggle)

            titleBarLayout.insertWidget(0, self._collapseTrigger)

            addStyleClass(self._collapseTrigger, 'group-box__collapse-trigger')
            addStyleClass(self._titleBar, 'group-box__title-bar--collapsable')

            self.collapse(collapsed)

    def setTitle(self, title):
        """Set the box title."""
        self._titleBarLabel.setText(title)

    def layout(self):
        """Return the box layout."""
        return self._contentBox.layout()

    def setLayout(self, layout):
        """Set the box layout."""
        self._contentBox.setLayout(layout)

    def addLayout(self, layout):
        """Add inner layout to the end of the main layout."""
        self.layout().addLayout(layout)

    def addWidget(self, widget):
        """Add widget to the end of the layout."""
        self.layout().addWidget(widget)

    def titleBar(self):
        """Return the widget for the title bar."""
        return self._titleBar

    def titleBarLabel(self):
        """Return the widget for the title bar label."""
        return self._titleBarLabel

    def buttonsBar(self):
        """Return the buttons bar."""
        return self._buttonsBarLayout

    def collapse(self, collapse=None):
        """Hide/show content."""
        if collapse or collapse is None:
            self._contentBox.hide()
        else:
            self._contentBox.show()

        self.updateGeometry()

        self._updateIcon()

    def isCollapsed(self):
        """Check if content is hidden."""
        return self._contentBox.isHidden()

    def _onCollapseToggle(self):
        collapsed = not self.isCollapsed()

        self.collapse(collapsed)

        self.collapseToggled.emit(collapsed)

    def _updateIcon(self):
        if self._contentBox.isHidden():
            icon = loadIcon('arrow_right')
        else:
            icon = loadIcon('arrow_down')

        self._collapseTrigger.setIcon(icon)


class GroupHBox(GroupBox):
    """Create a horizontal box area that can be collapsed."""

    def __init__(self, title='', icon=None, collapsable=True, collapsed=False):
        super().__init__(title, icon, collapsable, collapsed)

        layout = qtw.QHBoxLayout()

        self.setLayout(layout)


class GroupVBox(GroupBox):
    """Create a vertical box area that can be collapsed."""

    def __init__(self, title='', icon=None, collapsable=True, collapsed=False):
        super().__init__(title, icon, collapsable, collapsed)

        layout = qtw.QVBoxLayout()
        layout.setAlignment(qt.Qt.AlignTop)

        self.setLayout(layout)


# --------------------------------------------------------------------------------------------------


class ToolBar(qtw.QToolBar):
    """A movable panel that contains a set of controls."""

    class ButtonStyle(qtw.QProxyStyle):
        def styleHint(self, hint, *args, **kwargs):
            """Return an integer representing the specified style hint for the given widget."""
            if hint == qtw.QStyle.SH_ToolButton_PopupDelay:
                return 200

            return super().styleHint(hint, *args, **kwargs)

    def __init__(self, vertical: bool = False):
        super().__init__()

        addStyleClass(self, 'toolbar')

        if vertical:
            self.setOrientation(qt.Qt.Vertical)

        self._buttonStyle = None

    def addAction(
        self,
        icon: qtg.QIcon,
        text: str,
        receiver: t.Callable = None,
        menu: qtw.QMenu = None,
    ) -> qtw.QAction:
        """Add a button."""
        action = super().addAction(icon, text, receiver)
        actionIndex = len(self.actions()) - 1

        button = self.widgetForAction(action)
        addStyleClass(button, 'toolbar__button')

        if menu:
            action.setMenu(menu)

            if receiver:
                button.setPopupMode(qtw.QToolButton.DelayedPopup)
                addStyleClass(button, 'toolbar__button--with-menu')

                button.setContextMenuPolicy(qt.Qt.CustomContextMenu)
                button.customContextMenuRequested.connect(
                    partial(self._onButtonRightClick, actionIndex)
                )

                if not self._buttonStyle:
                    self._buttonStyle = ToolBar.ButtonStyle()

                button.setStyle(self._buttonStyle)
            else:
                button.setPopupMode(qtw.QToolButton.InstantPopup)

        return action

    def _onButtonRightClick(self, actionIndex, _pos):
        self.widgetForAction(self.actions()[actionIndex]).showMenu()


# --------------------------------------------------------------------------------------------------


class NodeField(qtw.QWidget):
    """Field to store the name of a maya node."""

    nodePicked = qt.Signal(str)

    def __init__(
        self,
        hasPickButton=True,
        placeholderText='None',
        onDisplayName=None,
        onSelect=None,
    ):
        super().__init__()

        addStyleClass(self, 'node-field')

        self._nodeName = None
        self._onDisplayName = onDisplayName
        self._onSelect = onSelect

        self._textField = ClickableLineEdit()
        self._textField.setReadOnly(True)
        self._textField.setPlaceholderText(placeholderText)
        self._textField.clicked.connect(self.selectNode)
        addStyleClass(self._textField, 'node-field__text-field')

        layout = qtw.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        layout.addWidget(self._textField)

        if hasPickButton:
            pickButton = qtw.QPushButton()
            pickButton.setIcon(loadIcon('add'))
            pickButton.setSizePolicy(qtw.QSizePolicy.Minimum, qtw.QSizePolicy.Minimum)
            pickButton.clicked.connect(self.pickNode)
            addStyleClass(pickButton, 'node-field__pick-button')

            addStyleClass(self, 'node-field--with-pick-button')

            layout.addWidget(pickButton)

        self.setLayout(layout)

    def setNode(self, nodeName):
        """Set the node name."""
        self._nodeName = nodeName

        if nodeName:
            try:
                displayName = mx.Node(nodeName).uniqueName
            except mx.MayaNodeError:
                displayName = nodeName
        else:
            displayName = ''

        if displayName and self._onDisplayName:
            displayName = self._onDisplayName(displayName)

        self._textField.setText(displayName)

    def pickNode(self):
        """Pick node name from selection."""
        selectionList = mx.cmd.ls(selection=True)

        if len(selectionList) > 1:
            log.warning('Too many objects selected.')
            return

        self._nodeName = selectionList[0].uniqueName if selectionList else ''

        displayName = self._onDisplayName(self._nodeName) if self._onDisplayName else self._nodeName

        self._textField.setText(displayName)

        self.nodePicked.emit(self._nodeName)

    def selectNode(self):
        """Select the node."""
        if self._nodeName and (not self._onSelect or self._onSelect(self._nodeName)):
            try:
                mx.cmd.select(self._nodeName)
            except ValueError as error:
                log.error(str(error))


class ClickableLineEdit(qtw.QLineEdit):
    """Line edit that emits mouse events."""

    clicked = qt.Signal()

    def __init__(self, contents=None, parent=None):
        super().__init__(contents, parent)

        self._mouseMoved = False

    def mousePressEvent(self, event):
        """Receive mouse press event."""
        super().mousePressEvent(event)

        self._mouseMoved = False

    def mouseReleaseEvent(self, event):
        """Receive mouse release event."""
        super().mouseReleaseEvent(event)

        if not self._mouseMoved:
            self.clicked.emit()

    def mouseMoveEvent(self, event):
        """Receive mouse move event."""
        super().mouseMoveEvent(event)

        self._mouseMoved = True


# --------------------------------------------------------------------------------------------------


class Menu(qtw.QMenu):
    """A menu with customizable items."""

    def __init__(self, titleOrParent=None, parent=None):
        super().__init__(titleOrParent, parent)

        addStyleClass(self, 'menu')

        self._previousHoveredAction = None

        self.aboutToHide.connect(self._onAboutToHide)
        self.hovered.connect(self._onActionHovered)

    def addAction(self, text, receiver=None, icon=None):
        """Add a new action to the menu."""
        action = MenuAction(self)

        action.setText(text)

        if icon:
            action.setIcon(icon)

        if receiver:
            action.triggered.connect(receiver)

        super().addAction(action)

        return action

    def addSection(self, text):
        """Add a new section action."""
        addStyleClass(self, 'menu--with-section')

        action = self._createSeparatorAction(text)

        super().addAction(action)

        return action

    def addSeparator(self):
        """Add a new separator action."""
        action = self._createSeparatorAction()

        super().addAction(action)

        return action

    def addMenu(self, menuOrTitle, icon=None):
        """Add a submenu."""
        if isinstance(menuOrTitle, str):
            menu = Menu(menuOrTitle, self)
        else:
            menu = menuOrTitle

        action = MenuAction(self)
        action.setText(menu.title())
        action.setMenu(menu)

        if icon:
            action.setIcon(icon)

        super().addAction(action)

        if isinstance(menuOrTitle, str):
            return menu

        return action

    def _createSeparatorAction(self, text=None):
        section = qtw.QWidget()
        addStyleClass(section, 'menu-section')

        layout = qtw.QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        section.setLayout(layout)

        if text:
            label = qtw.QLabel(text)
            addStyleClass(label, 'menu-section__label')
            addStyleClass(section, 'menu-section--with-label')
            layout.addWidget(label)

        separator = qtw.QLabel()
        addStyleClass(separator, 'menu-section__separator')
        layout.addWidget(separator, 1)

        action = qtw.QWidgetAction(self)
        action.setSeparator(True)
        action.setDefaultWidget(section)

        return action

    def _onAboutToHide(self):
        self._previousHoveredAction = None

    def _onActionHovered(self, action):
        if action == self._previousHoveredAction:
            return

        menu = action.parentWidget()
        submenu = action.menu()

        actionsToUnhighlight = menu.actions() + (submenu.actions() if submenu else [])

        for siblingAction in actionsToUnhighlight:
            if isinstance(siblingAction, MenuAction):  # a section/separator is not a MenuAction
                siblingAction.menuItem().highlight(False)

        action.menuItem().highlight()

        if self._previousHoveredAction:
            # Manually hide the submenus to have the menu items activated faster.
            previousMenu = self._previousHoveredAction.parentWidget()
            previousSubmenu = self._previousHoveredAction.menu()

            if previousMenu != menu and previousMenu != submenu and previousSubmenu != menu:
                previousMenu.hide()

            if previousSubmenu and previousSubmenu != menu:
                previousSubmenu.hide()

        self._previousHoveredAction = action


class MenuItem(qtw.QWidget):
    """The widget for the menu item."""

    def __init__(self, action, parent):
        super().__init__(parent)

        self.setAttribute(qt.Qt.WA_StyledBackground)
        self.setMouseTracking(True)

        addStyleClass(self, 'menu-item')

        self._action = action
        action.changed.connect(self._onActionChanged)

        text = action.text()
        icon = action.icon()
        shortcut = action.shortcut().toString()

        self._checkboxWidget = qtw.QCheckBox()
        self._checkboxWidget.setAttribute(qt.Qt.WA_TransparentForMouseEvents)
        self._checkboxWidget.setChecked(action.isChecked())
        self._checkboxWidget.setVisible(action.isCheckable())
        addStyleClass(self._checkboxWidget, 'menu-item__checkbox')

        self._iconWidget = qtw.QLabel()
        self._iconWidget.setAttribute(qt.Qt.WA_TransparentForMouseEvents)
        self._iconWidget.setAlignment(qt.Qt.AlignCenter)
        addStyleClass(self._iconWidget, 'menu-item__icon')
        if icon:
            self._iconWidget.setPixmap(icon.pixmap(16 * SCALE_FACTOR, 16 * SCALE_FACTOR))

        self._labelWidget = qtw.QLabel(text)
        self._labelWidget.setAttribute(qt.Qt.WA_TransparentForMouseEvents)
        addStyleClass(self._labelWidget, 'menu-item__label')

        self._shortcutWidget = qtw.QLabel(shortcut)
        self._shortcutWidget.setAttribute(qt.Qt.WA_TransparentForMouseEvents)
        addStyleClass(self._shortcutWidget, 'menu-item__shortcut')

        layout = qtw.QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._iconWidget)
        layout.addWidget(self._checkboxWidget)
        layout.addWidget(self._labelWidget, 1)
        layout.addWidget(self._shortcutWidget)

        if action.menu():
            submenuArrow = qtw.QLabel()
            submenuArrow.setAttribute(qt.Qt.WA_TransparentForMouseEvents)
            submenuArrow.setPixmap(
                loadIcon('arrow_right').pixmap(13 * SCALE_FACTOR, 13 * SCALE_FACTOR)
            )
            addStyleClass(submenuArrow, 'menu-item__submenu-arrow')
            layout.addWidget(submenuArrow)

        self.setLayout(layout)

    def addWidget(self, widget):
        """Append a widget to the end of the menu item."""
        if isinstance(widget, qtw.QPushButton):
            widget.clicked.connect(self.parentWidget().hide)
            addStyleClass(widget, 'menu-item__button')

        self.layout().addWidget(widget)

    def highlight(self, highlight=True):
        """Highlight the menu item."""
        if highlight:
            addStyleClass(self, 'menu-item--active')
        else:
            removeStyleClass(self, 'menu-item--active')

        icon = self._action.icon()

        if icon:
            if highlight and self._action.isEnabled():
                iconMode = qtg.QIcon.Active
            else:
                iconMode = qtg.QIcon.Normal

            self._iconWidget.setPixmap(icon.pixmap(16 * SCALE_FACTOR, 16 * SCALE_FACTOR, iconMode))

    def enterEvent(self, event):
        """Override to highlight the menu item on mouse enter."""
        super().enterEvent(event)

        self.highlight()

    def leaveEvent(self, event):
        """Override to unhighlight the menu item on mouse leave."""
        super().leaveEvent(event)

        self.highlight(False)

    def _onActionChanged(self):
        action = self._action

        self._shortcutWidget.setText(action.shortcut().toString())

        if action.isCheckable():
            self._checkboxWidget.setChecked(action.isChecked())
            self._checkboxWidget.show()
            self._iconWidget.hide()
        else:
            self._checkboxWidget.hide()
            self._iconWidget.show()


class MenuAction(qtw.QWidgetAction):
    """The action of the menu item."""

    def __init__(self, parent):
        super().__init__(parent)

        self._menuItem = None

    def createWidget(self, parent):
        """Create the menu item."""
        if self._menuItem:
            oldParent = self._menuItem.parent()

            if parent != oldParent and isinstance(oldParent, Menu):
                # This is necessary since the tearoff menu puts all items in a new widget.
                addStyleClass(parent, getStyleClass(oldParent))

            self._menuItem.setParent(parent)

            return self._menuItem

        self._menuItem = MenuItem(self, parent)

        return self._menuItem

    def deleteWidget(self, widget):
        """Delete the menu item."""
        super().deleteWidget(widget)

        self._menuItem = None

    def menuItem(self):
        """Return the menu item."""
        return self._menuItem


class MenuButton(qtw.QWidget):
    """A push button that opens a menu when clicked."""

    defaultActionClicked = qt.Signal()
    menuAboutToShow = qt.Signal(qtw.QMenu)

    def __init__(self):
        super().__init__()

        addStyleClass(self, 'menu-button')

        self._menu = Menu(self)
        self._menu.aboutToShow.connect(lambda: self.menuAboutToShow.emit(self._menu))

        triggerBtn = qtw.QPushButton()
        triggerBtn.setMenu(self._menu)
        triggerBtn.setIcon(loadIcon('menu'))
        addStyleClass(triggerBtn, 'menu-button__button menu-button__trigger')

        layout = qtw.QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(triggerBtn)

        self.setLayout(layout)

    def menu(self):
        """Return the menu."""
        return self._menu

    def setDefaultAction(self, icon=None, text=''):
        """Enable the default action button."""
        actionBtn = qtw.QPushButton(text)
        actionBtn.clicked.connect(self.defaultActionClicked.emit)

        if icon:
            actionBtn.setIcon(icon)

        addStyleClass(actionBtn, 'menu-button__button')
        addStyleClass(self, 'menu-button--with-default-action')

        self.layout().insertWidget(0, actionBtn, 1)


class SubmenuCreator:
    """Utility to create all intermediate-level submenus needed for a given path."""

    def __init__(self, menu: qtw.QMenu, pathSeparator='|'):
        self.menu = menu
        self.pathSeparator = pathSeparator
        self.submenus = {}

    def __call__(
        self,
        path: str,
        onCreation: t.Callable[[qtw.QMenu, str], None] = None,
    ) -> qtw.QMenu:
        """Create a submenu for the given path."""
        activeMenu = self.menu
        isRootPath = path.startswith(self.pathSeparator)

        if isRootPath:
            groups = path.split(self.pathSeparator)[1:-1]
        else:
            groups = path.split(self.pathSeparator)[:-1]

        for i, grp in enumerate(groups):
            groupPath = self.pathSeparator.join(groups[: i + 1])

            if groupPath not in self.submenus:
                self.submenus[groupPath] = activeMenu.addMenu(grp)

                if onCreation:
                    activePath = (self.pathSeparator if isRootPath else '') + groupPath

                    onCreation(self.submenus[groupPath], activePath)

            activeMenu = self.submenus[groupPath]

        return activeMenu


# --------------------------------------------------------------------------------------------------


class ColorField(qtw.QPushButton):
    """Color field."""

    colorPicked = qt.Signal(tuple)

    def __init__(self, rgb=(0, 0, 0), renderingSpace=False):
        super().__init__()

        addStyleClass(self, 'color-field')

        self._rgb = rgb
        self._renderingSpace = renderingSpace

        self.setColor(rgb)

        self.setSizePolicy(qtw.QSizePolicy.Expanding, qtw.QSizePolicy.Fixed)

        self.clicked.connect(self._showColorPicker)

    def color(self):
        """Get the color as RGB."""
        return self._rgb

    def setColor(self, rgb):
        """Set the color's RGB values."""
        self._rgb = tuple(rgb)

        displayColor = qtg.QColor()

        if self._renderingSpace:
            displayRgb = mx.cmd.colorManagementConvert(toDisplaySpace=rgb)

            displayRgb = [round(value, 3) for value in displayRgb]
            displayRgb = [min(max(0, value), 1) for value in displayRgb]

            displayColor.setRgbF(*displayRgb)
        else:
            displayColor.setRgb(*rgb)

        self.setStyleSheet(f'background: {displayColor.name()}')

    def _showColorPicker(self):
        if self._renderingSpace:
            mx.cmd.colorEditor(rgb=self._rgb)

            if not mx.cmd.colorEditor(query=True, result=True):
                return

            rgb = mx.cmd.colorEditor(query=True, rgb=True)
        else:
            selectedColor = qtw.QColorDialog.getColor(qtg.QColor(*self._rgb), self.parent())

            if not selectedColor.isValid():
                return

            rgb = selectedColor.getRgb()[:3]  # remove alpha

        self.setColor(rgb)

        self.colorPicked.emit(self._rgb)


class ToggleButton(qtw.QWidget):
    """Toggle button with multi-state."""

    stateChanged = qt.Signal(object)

    def __init__(self):
        super().__init__()

        self._buttons = []
        self._values = []

        self._activeBtn = None
        self._activeValue = None

        addStyleClass(self, 'toggle-button')

        layout = qtw.QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def addState(self, label: str, value: t.Any) -> qtw.QPushButton:
        """Add a state."""
        btn = qtw.QPushButton(label)
        btn.clicked.connect(lambda: self._onStateChange(value))

        addStyleClass(btn, 'toggle-button__button')

        if not self._buttons:
            addStyleClass(btn, 'toggle-button__button--first')

        if self._buttons:
            removeStyleClass(self._buttons[-1], 'toggle-button__button--last')

        if len(self._buttons) > 1:
            addStyleClass(self._buttons[-1], 'toggle-button__button--middle')

        addStyleClass(btn, 'toggle-button__button--last')

        self._buttons.append(btn)
        self._values.append(value)

        self.layout().addWidget(btn)

        if not self._activeValue:
            self.setState(value)

        return btn

    def state(self) -> t.Any:
        """Get active state."""
        return self._activeValue

    def setState(self, value: t.Any):
        """Set active state."""
        try:
            btn = self._buttons[self._values.index(value)]
        except ValueError as error:
            raise Exception(f'Toggle Button: Invalid state value: {value}') from error

        for stateBtn in self._buttons:
            removeStyleClass(stateBtn, 'toggle-button__button--active')

        addStyleClass(btn, 'toggle-button__button--active')

        self._activeBtn = btn
        self._activeValue = value

    def _onStateChange(self, value):
        self.setState(value)

        self.stateChanged.emit(value)


class Slider(qtw.QWidget):
    """A slider with more options."""

    valueChanged = qt.Signal(float)

    def __init__(self, minValue: float | int, maxValue: float | int):
        super().__init__()

        addStyleClass(self, 'slider')

        self._min = minValue
        self._max = maxValue
        self._value = minValue
        self._isFloat = isinstance(minValue, float) or isinstance(maxValue, float)

        self._slider = qtw.QSlider(qt.Qt.Horizontal)

        if self._isFloat:
            self._slider.setMinimum(0)
            self._slider.setMaximum(100)
            self._slider.setValue(0)
        else:
            self._slider.setMinimum(minValue)
            self._slider.setMaximum(maxValue)
            self._slider.setValue(minValue)

        self._slider.valueChanged.connect(self._onSliderValueChanged)

        self._textField = qtw.QLineEdit(str(self._value))

        if self._isFloat:
            textFieldValidator = qtg.QDoubleValidator(minValue, maxValue, 3)
            textFieldValidator.setNotation(qtg.QDoubleValidator.StandardNotation)
        else:
            textFieldValidator = qtg.QIntValidator(minValue, maxValue)

        self._textField.setValidator(textFieldValidator)
        self._textField.editingFinished.connect(self._onTextFieldChanged)

        layout = qtw.QHBoxLayout()
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(self._textField, 1)
        layout.addWidget(self._slider, 3)

        self.setLayout(layout)

    def setValue(self, value):
        """Change the current value."""
        self._value = value

        if self._isFloat:
            sliderValue = (value - self._min) / (self._max - self._min) * 100
        else:
            sliderValue = value

        self._slider.setValue(sliderValue)

    def _onSliderValueChanged(self, value):
        if self._isFloat:
            self._value = round(self._min + (self._max - self._min) * (value / 100), 3)
        else:
            self._value = value

        self._textField.setText(str(self._value))

        self.valueChanged.emit(self._value)

    def _onTextFieldChanged(self):
        value = self._textField.text()
        value = float(value) if self._isFloat else int(value)

        self.setValue(value)


class VectorField(qtw.QWidget):
    """A widget holding the values of a vector."""

    valueChanged = qt.Signal(tuple)

    def __init__(self):
        super().__init__()

        addStyleClass(self, 'vector-field')

        self._xField = qtw.QLineEdit('0.0')
        self._yField = qtw.QLineEdit('0.0')
        self._zField = qtw.QLineEdit('0.0')

        textFieldValidator = qtg.QDoubleValidator()
        textFieldValidator.setDecimals(3)
        textFieldValidator.setNotation(qtg.QDoubleValidator.StandardNotation)

        self._xField.setValidator(textFieldValidator)
        self._yField.setValidator(textFieldValidator)
        self._zField.setValidator(textFieldValidator)

        self._xField.textEdited.connect(self._onValueChanged)
        self._yField.textEdited.connect(self._onValueChanged)
        self._zField.textEdited.connect(self._onValueChanged)

        layout = qtw.QHBoxLayout()
        layout.setSpacing(6)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(self._xField)
        layout.addWidget(self._yField)
        layout.addWidget(self._zField)

        self.setLayout(layout)

    def value(self) -> tuple:
        """Retrieve the vector values."""
        try:
            xValue = float(self._xField.text())
        except ValueError:
            xValue = 0.0

        try:
            yValue = float(self._yField.text())
        except ValueError:
            yValue = 0.0

        try:
            zValue = float(self._zField.text())
        except ValueError:
            zValue = 0.0

        return (xValue, yValue, zValue)

    def setValue(self, value: tuple):
        """Set the vector values."""
        self._xField.setText(str(round(value[0], 3)))
        self._yField.setText(str(round(value[1], 3)))
        self._zField.setText(str(round(value[2], 3)))

    def _onValueChanged(self):
        self.valueChanged.emit(self.value())
