# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Utiliy functions."""

import os
import sys
import re
import functools
import weakref
from importlib import import_module

import PySide2.QtCore as qt


# --------------------------------------------------------------------------------------------------
# Functions
# --------------------------------------------------------------------------------------------------


def getClassByFullName(fullName):
    """Get a class from a string.

    Parameters
    ----------
    fullName : str
        The full name of the class, e.g. ``puppeteer.meta.MetaNode``.

    Returns
    -------
    type
        The class.
    """
    moduleName, _, className = fullName.rpartition('.')
    module = import_module(moduleName)
    return getattr(module, className)


def getClassByType(cls, module):
    """Get a list of subclasses from a module.

    Parameters
    ----------
    module : str or module
        The module to import from.

    Returns
    -------
    list
        A list of classes.
    """
    if isinstance(module, str):
        module = import_module(module)

    classes = []

    for attrName in dir(module):
        attr = getattr(module, attrName)
        try:
            if issubclass(attr, cls):
                classes.append(attr)
        except TypeError:
            pass

    return classes


def getClassName(cls):
    """Get the name of a class."""
    return cls.__name__


def getClassFullName(cls):
    """Get the full name of a class."""
    return f'{cls.__module__}.{cls.__name__}'


def getClassFriendlyName(cls):
    """Get the class name as a human-friendly string."""
    return getFriendlyName(getClassName(cls))


def getFriendlyName(name):
    """Get a human-friendly string from a camel/pascal case identifier."""
    friendlyName = re.sub(
        r'((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))',
        r' \1',
        name,
    )
    return friendlyName[0].upper() + friendlyName[1:]


def findFiles(
    path,  # pylint: disable=too-many-arguments
    nameFilters=(),
    filesToIgnore=(),
    returnFilesOnly=False,
    returnFileName=False,
    returnNoExtension=False,
    returnNoBasePath=False,
):
    """Find all the files and directories from a specified path.

    Parameters
    ----------
    path : str
        The base file path.
    nameFilters : list[str]
        A wildcard (globbing) filter that understands * and ? wildcards. See `QtCore.QDir`.
    filesToIgnore : list[str]
        A list of files to ignore.
    returnFilesOnly : bool
        True to not return directories.
    returnFileName : bool
        True to return only the base file name.
    returnNoExtension : bool
        True to exclude the file extention.
    returnNoBasePath : bool
        True to exclude the provided `path`.

    Returns
    -------
    list
        The list of found entries.
    """
    dirFilter = qt.QDir.NoDotAndDotDot | qt.QDir.Files
    if not returnFilesOnly:
        dirFilter |= qt.QDir.AllDirs

    qDir = qt.QDir(path)
    qDir.setFilter(dirFilter)

    if nameFilters:
        qDir.setNameFilters(nameFilters)

    files = []

    dirIt = qt.QDirIterator(qDir, qt.QDirIterator.Subdirectories)
    while dirIt.hasNext():
        filePath = dirIt.next()
        fileName = dirIt.fileName()

        if filesToIgnore and fileName in filesToIgnore:
            continue

        foundFile = fileName if returnFileName else filePath

        if returnNoExtension:
            foundFile = foundFile.rsplit('.', 1)[0]

        if returnNoBasePath:
            foundFile = foundFile.replace(qDir.path(), '')[1:]

        files.append(foundFile)

    return files


def findPyModules(path, basePackage):
    """Find a list of python modules.

    Parameters
    ----------
    path : str
        The base path to look into.
    basePackage : str
        The package from where to import the modules.

    Returns
    -------
    list
        A list of modules.
    """
    files = findFiles(
        path,
        nameFilters=['*.py'],
        filesToIgnore=['__init__.py'],
        returnFilesOnly=True,
        returnNoExtension=True,
        returnNoBasePath=True,
    )

    modules = []

    for filePath in files:
        name = os.path.normpath(filePath).replace(os.path.sep, '.')
        name = f'{basePackage}.{name}'
        modules.append(import_module(name))

    return modules


def removePyModules(basePackage, ignoredModules=()):
    """Remove modules from ``sys.modules``.

    Parameters
    ----------
    basePackage : str
        The base package from which to remove the modules.
    ignoredModules : list[str]
        A list of modules to ignore.
        The module names should contain the `basePackage`.
    """
    modulesToDelete = []

    for module in sys.modules:
        if module in ignoredModules:
            continue

        if module.startswith(basePackage):
            modulesToDelete.append(module)

    for module in modulesToDelete:
        del sys.modules[module]


def weakMethodProxy(method):
    """Return a weak reference proxy to a method."""
    instance = method.__self__
    unboundMethod = getattr(instance.__class__, method.__name__)

    return functools.partial(unboundMethod, weakref.proxy(instance))
